import inspect
import asyncio
import atexit
from pygears.core.err import register_exit_hook
from pygears import registry, GearDone
from pygears.sim import clk, timestep, delta, sim_log, sim_phase
from pygears.sim.sim import cancel
from pygears.typing_common.codec import code
from pygears.typing import typeof, TLM


def is_async_gen(func):
    return bool(func.__code__.co_flags & inspect.CO_ASYNC_GENERATOR)


def is_simgear_func(func):
    return inspect.iscoroutinefunction(func) or is_async_gen(func)


class SimGear:
    def __init__(self, gear):
        self.gear = gear
        self.out_queues = []
        self.namespace = registry('SimMap')
        self.done = False
        self._clean = True
        if not hasattr(self, 'func'):
            self.func = gear.func

    @property
    def sim_func_args(self):
        args = []
        for p in self.gear.in_ports:
            args.append(p.consumer)

        kwds = {
            k: self.gear.params[k]
            for k in self.gear.kwdnames if k in self.gear.params
        }

        return args, kwds

    def finish(self):
        self.done = True
        # self.task.cancel()
        for port in self.gear.out_ports:
            port.producer.finish()

        if not self._clean:
            self.cleanup()

    def cleanup(self):
        self._clean = True

    def setup(self):
        self._clean = False
        atexit.register(self.finish)
        register_exit_hook(self.cleanup)
        if self.gear.params['sim_setup'] is not None:
            self.gear.params['sim_setup'](self.gear)

    async def run(self):
        self.task = asyncio.Task.current_task()
        args, kwds = self.sim_func_args

        # out_ports = self.gear.out_ports
        # if len(out_ports) == 1:
        #     out_ports = out_ports[0]

        out_prods = [p.producer for p in self.gear.out_ports]
        single_output = len(out_prods) == 1
        if single_output:
            out_prods = out_prods[0]

        sim = registry('Simulator')

        try:
            if is_async_gen(self.func):
                while(1):
                    if sim.phase != 'forward':
                        await clk()

                    async for val in self.func(*args, **kwds):
                        if sim.phase != 'forward':
                            await clk()

                        if val is not None:
                            if single_output:
                                out_prods.put_nb(val)
                                await out_prods.ready()
                            else:
                                for p, v in zip(out_prods, val):
                                    if v is not None:
                                        p.put_nb(v)

                                for p, v in zip(out_prods, val):
                                    if v is not None:
                                        await p.ready()

                    if args:
                        if all(a.done() for a in args):
                            raise GearDone
            else:
                while (1):
                    await self.func(*args, **kwds)

                    if args:
                        if all(a.done() for a in args):
                            raise GearDone

        except GearDone as e:
            for p in self.gear.in_ports:
                intf = p.consumer
                if not intf.empty():
                    prod_intf = intf.in_queue.intf
                    prod_gear = prod_intf.consumers[0].gear
                    cancel(prod_gear)

            self.finish()
            raise e



from pygears.core.gear import gear
from pygears.util.utils import gather


@gear
async def ccat(*din) -> b'Tuple[din]':
    async with gather(*din) as dout:
        yield dout

    # din_data = []
    # for d in din:
    #     din_data.append(await d.pull())

    # yield tuple(din_data)

    # for d in din:
    #     d.ack()



from pygears.svgen.svmod import SVModuleGen
from pygears.typing.queue import Queue
from pygears.svgen.inst import SVGenInstPlugin
from pygears.common.cart import cart, cart_sync
from .syncguard import SVGenSyncGuard
from .cat_util import din_data_cat


class SVGenCartBase(SVModuleGen):
    @property
    def is_generated(self):
        return True

    def get_sv_port_config(self, modport, type_, name):
        cfg = super().get_sv_port_config(modport, type_, name)

        if issubclass(type_, Queue):
            cfg['lvl'] = type_.lvl
            cfg['data_eot'] = f'(&{name}_s.eot)'
        else:
            cfg['lvl'] = 0
            cfg['data_eot'] = 1

        return cfg


class SVGenCart(SVGenCartBase):
    @property
    def is_generated(self):
        return True

    def get_module(self, template_env):
        intfs = list(self.sv_port_configs())
        queue_intfs = [
            i for i in intfs if i['lvl'] > 0 and i['modport'] == 'consumer'
        ]

        context = {
            'queue_intfs': queue_intfs,
            'module_name': self.sv_module_name,
            'intfs': intfs,
            'din_data_cat': din_data_cat
        }

        return template_env.render_local(__file__, "cart.j2", context)


class SVGenCartSyncBase(SVGenCartBase):
    def __init__(self, node):
        super().__init__(node)

        if 'outsync' not in self.node.params:
            self.node.params['outsync'] = True

        if self.node.params['outsync']:
            self.syncguard = SVGenSyncGuard(f'{self.sv_module_name}_syncguard',
                                            len(self.node.in_ports))
        else:
            self.syncguard = None

    @property
    def sv_file_name(self):
        if self.syncguard is None:
            return super().sv_file_name
        else:
            return super().sv_file_name, self.syncguard.sv_file_name

    def get_module(self, template_env, template_fn):

        context = {
            'outsync': self.node.params['outsync'],
            'module_name': self.sv_module_name,
            'intfs': list(self.sv_port_configs()),
            'din_data_cat': din_data_cat
        }
        contents = template_env.render_local(__file__, template_fn, context)

        if self.syncguard is None:
            return contents
        else:
            return contents, self.syncguard.get_module(template_env)


class SVGenCartSync(SVGenCartSyncBase):
    def get_module(self, template_env):

        return super().get_module(template_env, 'cart_sync.j2')


class SVGenCartPlugin(SVGenInstPlugin):
    @classmethod
    def bind(cls):
        cls.registry['SVGenModuleNamespace'][cart_sync] = SVGenCartSync
        cls.registry['SVGenModuleNamespace'][cart] = SVGenCart



from pygears.typing.base import TypingMeta


def customize(dtype, val_map, name_map={}):
    map_dict = {n: val_map[name_map.get(n, n)] for n in dtype.templates}

    return dtype[map_dict]


def repack(dtype, val, name_map={}, val_map={}):
    mapped_names = [name_map.get(f, f) for f in dtype.fields]

    map_dict = {n: val.get(n, val_map.get(n, None)) for n in mapped_names}

    if isinstance(val, TypingMeta):
        return dtype.base[map_dict]
    else:
        return dtype(map_dict)



from pygears import gear, QueueEmpty, module
from pygears.sim import clk


def setup(module):
    module.reg = None


@gear(sim_setup=setup, svgen={'svmod_fn': 'dreg.sv'})
async def dreg(din: 'tdin') -> b'tdin':
    if module().reg is None:
        module().reg = await din.get()
    else:

        yield module().reg

        try:
            module().reg = din.get_nb()
        except QueueEmpty:
            module().reg = None



from pygears.svgen.inst import SVGenInstPlugin
from pygears.common.ccat import ccat
from pygears.svgen.svmod import SVModuleGen


class SVGenCCat(SVModuleGen):
    @property
    def is_generated(self):
        return True

    def get_module(self, template_env):
        context = {
            'module_name': self.sv_module_name,
            'intfs': list(self.sv_port_configs())
        }

        return template_env.render_local(__file__, "ccat.j2", context)


class SVGenCCatPlugin(SVGenInstPlugin):
    @classmethod
    def bind(cls):
        cls.registry['SVGenModuleNamespace'][ccat] = SVGenCCat



from pygears.typing import Integer, Tuple, Queue, Int, typeof, Uint
from pygears import gear, alternative
from pygears.common import ccat, fmap, cart, permuted_apply
from pygears import module
from pygears.util.utils import quiter

TCfg = Tuple[{
    'start': Integer['w_start'],
    'cnt': Integer['w_cnt'],
    'incr': Integer['w_incr']
}]


def rng_out_type(cfg, cnt_steps):
    if cnt_steps:
        return cfg[0] + cfg[1] + cfg[2]
    else:
        return max(cfg[0], cfg[1])


@gear(svgen={'svmod_fn': 'rng.sv'})
async def sv_rng(cfg: TCfg,
                 *,
                 signed=b'typeof(cfg[0], Int)',
                 cnt_one_more=False,
                 cnt_steps=False,
                 incr_steps=False) -> Queue['rng_out_type(cfg, cnt_steps)']:
    def sign(x):
        return -1 if x < 0 else 1

    async with cfg as (start, cnt, incr):

        if not cnt_steps:
            rng_cfg = [int(start), int(cnt), int(incr)]
        else:
            rng_cfg = [
                int(start),
                int(start) + int(cnt) * int(incr),
                int(incr)
            ]

        rng_cfg[1] += sign(int(incr)) * cnt_one_more

        for data, last in quiter(range(*rng_cfg)):
            yield module().tout((data, last))


@gear
def rng(cfg: TCfg, *, cnt_steps=False, incr_steps=False, cnt_one_more=False):

    any_signed = any([typeof(d, Int) for d in cfg.dtype])
    all_signed = all([typeof(d, Int) for d in cfg.dtype])
    if any_signed and not all_signed:
        cfg = cfg | fmap(f=(Int, ) * len(cfg.dtype))

    return cfg | sv_rng(
        signed=any_signed,
        cnt_steps=cnt_steps,
        incr_steps=incr_steps,
        cnt_one_more=cnt_one_more)


@alternative(rng)
@gear
def rng_cnt_only(cfg: Integer['w_cnt'],
                 *,
                 cnt_steps=False,
                 incr_steps=False,
                 cnt_one_more=False):
    return ccat(0, cfg, 1) | rng


@alternative(rng)
@gear(enablement=b'len(cfg) == lvl')
def rng_multi_lvl(cfg: Tuple,
                  *,
                  lvl=1,
                  cnt_steps=False,
                  incr_steps=False,
                  cnt_one_more=False):
    return cfg \
        | fmap(name='forx', f=(None, rng(cnt_steps=cnt_steps,
                                         incr_steps=incr_steps,
                                         cnt_one_more=cnt_one_more)), fcat=cart) \
        | fmap(name='fory',
               f=fmap(
                   f=(rng(cnt_steps=cnt_steps,
                          incr_steps=incr_steps,
                          cnt_one_more=cnt_one_more), None),
                   fcat=permuted_apply(f=cart, indices=(1,0)))
                   # f=(rng, None), fcat=cart)
               )

    # | fmap(name='fory', f=(rng, None), lvl=2, fcat=permuted_apply(f=cart, indices=(1,0)))



import inspect

from pygears.typing.uint import Uint

from .base import EnumerableGenericMeta, type_str
from .bitw import bitw
from .unit import Unit


class UnionMeta(EnumerableGenericMeta):
    def __new__(cls, name, bases, namespace, args=[]):
        cls = super().__new__(cls, name, bases, namespace, args)

        args = cls.args
        if not args:
            return cls
        else:
            # flat_params = []
            # for a in args:
            #     if inspect.isclass(a) and issubclass(a, Union):
            #         flat_params.extend(a.args)
            #     else:
            #         flat_params.append(a)

            if len(args) == 0:
                return Unit
            elif len(args) == 1:
                return args[0]
            else:
                # cls.args = flat_params
                return cls

    def __getitem__(self, parameters):
        if not self.is_specified():
            return super().__getitem__(parameters)

        index = self.index_norm(parameters)

        if isinstance(index[0], slice):
            if(index[0].start == 0) and (index[0].stop == 1):
                return Uint[max(map(int, self.args))]
            elif(index[0].start == 1) and (index[0].stop == 2):
                return Uint[bitw(len(self.args) - 1)]
            elif(index[0].start == 0) and (index[0].stop == 2):
                return Uint[max(map(int, self.args))], Uint[bitw(len(self.args) - 1)]
            else:
                raise IndexError
        else:
            if (index[0] == 0):
                return Uint[max(map(int, self.args))]
            elif (index[0] == 1):
                return Uint[bitw(len(self.args) - 1)]
            else:
                raise IndexError

    def keys(self):
        return [0, 1]

    @property
    def types(self):
        return self.args

    def __str__(self):
        return '%s' % ' | '.join([type_str(a) for a in self.args])


class Union(tuple, metaclass=UnionMeta):
    def __new__(cls, val: tuple):
        return super(Union, cls).__new__(cls, (cls[0](val[0]), val[1]))

    # def __getitem__(self, index):
    #     index = type(self).index_norm(index)
    #     if len(index) == 2:
    #         return self
    #     else:
    #         index = index[0]
    #         val = super().__getitem__(index)

    #         if index == 0:
    #             type(self).decode(val)
    #             select = super().__getitem__(1)
    #             return type(self).types[select].decode(val)
    #         else:
    #             return val

    @property
    def data(self):
        return type(self).types[self[1]].decode(self[0])

    @classmethod
    def decode(cls, val):
        ret = []
        for t in cls:
            ret.append(t.decode(val))
            val >>= int(t)

        return cls(tuple(ret))



from pygears import GearDone, gear
from pygears.typing import TLM


@gear
async def seqr(*, t, seq) -> TLM['t']:
    for val in seq:
        yield val

    raise GearDone


@gear
async def delay_seqr(din: TLM['t'], *, seq) -> TLM['t']:
    item = await din.get()
    for val in seq:
        yield val

    # raise GearDone



from pygears import gear, Intf
from pygears.common import czip
from pygears.typing import Tuple, Uint, Union, Queue
from pygears.common import fmap, demux, decoupler, fifo, union_collapse
from pygears.cookbook import priority_mux, replicate

TCfg = Tuple[{'reduce_size': Uint['w_reduce_size'], 'init': 't_acc'}]


@gear
def reduce2(din, cfg: TCfg, *, f, max_size):

    acctype = cfg.dtype['init']

    qtype = Queue[acctype, din.dtype.lvl - 1]

    temp_res = Intf(dtype=qtype)
    cfg_rep = cfg | replicate
    sec_opnd = (cfg_rep, temp_res) \
        | priority_mux \
        | fmap(f=union_collapse, fcat=czip, lvl=1)

    result = czip(din, sec_opnd) | decoupler | fmap(f=f, fcat=czip, lvl=2)
    acc, fin_res = result | Union[qtype, qtype] | demux
    acc | fifo(intfs=[temp_res], depth=max_size)

    return fin_res



from .svgen import svgen
from .inst import register_sv_paths
from .generate import svgen_generate

from pygears.registry import load_plugin_folder
import os
load_plugin_folder(os.path.join(os.path.dirname(__file__), 'modules'))


__all__ = ['svgen', 'svgen_generate', 'register_sv_paths']



from pygears.typing import TypingNamespacePlugin, Queue, Tuple, Union, typeof


def factor(type_):
    if typeof(type_, Union):
        for t in type_.types:
            if not typeof(t, Queue):
                return type_
        else:
            union_types = []
            for t in type_.types:
                union_types.append(t[0])
                lvl = t.lvl

            return Queue[Union[tuple(union_types)], lvl]
    elif typeof(type_, Tuple):
        if all(typeof(t, Queue) for t in type_):
            tuple_types = [t[0] for t in type_]
            return Queue[Tuple[tuple(tuple_types)], type_[0].lvl]
    else:
        return type_


class TypeFactorPlugin(TypingNamespacePlugin):
    @classmethod
    def bind(cls):
        cls.registry['TypeArithNamespace']['factor'] = factor



def bitw(num: int) -> int:
    accum, shifter = 0, 1
    while num >= shifter:
        shifter <<= 1
        accum += 1
    return accum


def ceil_pow2(num: int) -> int:
    return int(2**(bitw(num-1)))



from pygears.core.gear import gear
from pygears.svgen.inst import SVGenInstPlugin
from pygears.svgen.svmod import SVModuleGen


@gear
async def priority_mux_valve(*din) -> b'Union[din]':
    pass


class SVGenPriorityMuxValve(SVModuleGen):
    @property
    def is_generated(self):
        return True

    def get_module(self, template_env):
        context = {
            'module_name': self.sv_module_name,
            'intfs': list(self.sv_port_configs())
        }
        return template_env.render_local(__file__, "priority_mux_valve.j2",
                                         context)


class SVGenPriorityMuxValvePlugin(SVGenInstPlugin):
    @classmethod
    def bind(cls):
        cls.registry['SVGenModuleNamespace'][
            priority_mux_valve] = SVGenPriorityMuxValve



from pygears.core.hier_node import HierVisitorBase
from pygears.svgen.util import svgen_visitor
from pygears.rtl.intf import RTLIntf


def is_in_subbranch(root, node):
    if node is None:
        return True

    root_path = root.name.split('/')
    node_path = node.name.split('/')

    # if len(root_path) > len(node_path) + 1:
    if len(root_path) > len(node_path):
        return False

    for r, n in zip(root_path, node_path):
        if r != n:
            return False
    else:
        return True


@svgen_visitor
class RTLOutChannelVisitor(HierVisitorBase):
    def RTLNode(self, node):
        super().HierNode(node)
        if node.parent is None:
            return True

        # print(node.name)
        # if node.name == 'is_ifm_mover/do_if/fmap':
        #     import pdb
        #     pdb.set_trace()

        for p in node.out_ports:
            cons_intf = p.consumer
            consumers_at_same_level_or_sublevel = [
                cons_p in node.parent.out_ports or is_in_subbranch(cons_intf.parent, cons_p.node.parent)
                for cons_p in cons_intf.consumers
            ]
            if not all(consumers_at_same_level_or_sublevel) or (not cons_intf):
                # print(f'Node: {node.name}')
                # print(f'    {consumers_at_same_level_or_sublevel}')

                # import pdb
                # pdb.set_trace()
                # print("Here!")

                cons_intf.parent.child.remove(cons_intf)
                cons_intf.parent.parent.add_child(cons_intf)

                cons_intf.disconnect(p)
                node.parent.add_out_port(
                    p.basename, consumer=cons_intf, dtype=cons_intf.dtype)

                out_port = node.parent.out_ports[-1]
                cons_intf.producer = out_port

                local_cons = [
                    port for port, same_lvl in zip(
                        cons_intf.consumers,
                        consumers_at_same_level_or_sublevel) if same_lvl
                ]

                local_intf = RTLIntf(node.parent, cons_intf.dtype, producer=p)
                local_intf.connect(out_port)
                p.consumer = local_intf

                for port in local_cons:
                    cons_intf.consumers.remove(port)
                    local_intf.connect(port)

        return True


@svgen_visitor
class RTLChannelVisitor(HierVisitorBase):
    def RTLNode(self, node):
        for p in node.in_ports:
            prod_intf = p.producer

            parent = node.parent
            while parent is not None:
                if (prod_intf is not None and prod_intf.parent != parent
                        and (not parent.is_descendent(prod_intf.parent))):

                    parent.add_in_port(
                        p.basename, producer=prod_intf, dtype=prod_intf.dtype)
                    in_port = parent.in_ports[-1]

                    local_cons = [
                        port for port in prod_intf.consumers
                        if parent.is_descendent(port.node)
                    ]

                    local_intf = RTLIntf(
                        parent, prod_intf.dtype, producer=in_port)

                    for port in local_cons:
                        prod_intf.consumers.remove(port)
                        local_intf.connect(port)

                    prod_intf.connect(in_port)

                parent = parent.parent



from pygears.core.gear import alternative, gear
from pygears.typing import Uint
from pygears.util.hof import oper_tree


@gear(enablement=b'len(din) == 2')
def neq(*din) -> Uint[1]:
    pass


@alternative(neq)
@gear
def neq_vararg(*din, enablement=b'len(din) > 2') -> Uint[1]:
    return oper_tree(din, neq)



from pygears import gear, module, GearDone
from pygears.sim import delta, clk


@gear
async def fifo(din, *, depth=2, threshold=0, regout=False) -> b'din':
    '''For this implementation depth must be a power of 2'''

    data = []
    out_data = False
    dout = module().dout

    while (1):
        if len(data) <= threshold and din.done():
            raise GearDone

        # TODO: Make fifo work correctly in corner case when it is full, but
        # consumer is ready
        if len(data) < depth:
            if not din.empty():
                data.insert(0, din.get_nb())

        if len(data) > threshold and not out_data and dout.ready_nb():
            dout.put_nb(data[-1])
            out_data = True

        await delta()

        if out_data and dout.ready_nb():
            data.pop()
            out_data = False

        await clk()



from pygears.core.hier_node import HierVisitorBase


class RTLConnectVisitor(HierVisitorBase):
    def HierNode(self, rtl_gear):
        super().HierNode(rtl_gear)
        if hasattr(rtl_gear, 'connect'):
            rtl_gear.connect()


def rtl_connect(top, conf):
    v = RTLConnectVisitor()
    v.visit(top)
    # for rtl_gear in top.child:
    #     v.visit(rtl_gear)

    return top.node



from pygears import gear, module
from pygears.typing import Integer, Uint
from pygears.core.intf import IntfOperPlugin


@gear
async def shr(din: Integer,
              cfg: Uint['w_shamt'],
              *,
              signed=b'typeof(din, Int)') -> b'din':

    async with cfg as shamt:
        async with din as d:
            yield module().tout(d >> shamt)


class MulIntfOperPlugin(IntfOperPlugin):
    @classmethod
    def bind(cls):
        cls.registry['IntfOperNamespace']['__rshift__'] = shr



from .sim_extend import SimExtend
from pygears.sim import sim_log


class SimExecution(SimExtend):
    # def before_call_forward(self, sim, sim_gear):
    #     return True

    def after_call_forward(self, sim, sim_gear):
        sim_log().info(f'forward')
        return True

    # def before_call_back(self, sim, sim_gear):
    #     return True

    def after_call_back(self, sim, sim_gear):
        sim_log().info(f'back')
        return True



from pygears import gear
from pygears.typing import Queue
from .ccat import ccat
from .czip import czip


@gear
def queue_wrap_from(din, qdin, *, fcat=czip):
    cat_data = fcat(qdin, din)

    return ccat(cat_data['data'][1], cat_data['eot']) \
        | Queue[din.dtype, qdin.dtype.lvl]



import os
from shutil import copyfile


def copy_file(fn, outdir, src):

    if not os.path.exists(outdir):
        os.makedirs(outdir)

    dest = os.path.join(outdir, fn)
    copyfile(src, dest)

    return dest


def save_file(fn, outdir, content):

    if not os.path.exists(outdir):
        os.makedirs(outdir)

    dest = os.path.join(outdir, fn)
    with open(os.path.join(outdir, fn), 'w') as content_file:
        content_file.write(content)

    return dest


def save_if_changed(fn, outdir, content):
    content_old = ''
    file_path = os.path.join(outdir, fn)
    if os.path.exists(file_path):
        with open(file_path) as f:
            content_old = f.read()

    if content_old != content:
        save_file(os.path.basename(fn), outdir, content)



from functools import wraps


def doublewrap(f):
    '''
    a decorator decorator, allowing the decorator to be used as:
    @decorator(with, arguments, and=kwds)
    or
    @decorator
    '''

    @wraps(f)
    def new_dec(*args, **kwds):
        if len(args) == 1 and len(kwds) == 0 and callable(args[0]):
            # actual decorated function
            return f(args[0])
        else:
            # decorator arguments
            return lambda realf: f(realf, *args, **kwds)

    return new_dec


def perpetum(func, *args, **kwds):
    while True:
        yield func(*args, **kwds)



from .accumulator import accumulator
from .replicate import replicate
from .chop import chop
from .clip import clip
from .din_cat import din_cat
from .form_sub_cfg import form_sub_cfg
from .iceil import iceil
from .priority_mux import priority_mux
from .qcnt import qcnt
from .qlen_cnt import qlen_cnt
from .release_after_eot import release_after_eot
from .reverse import reverse
from .rng import rng
from .sdp import sdp
from .take import take
from .tr_cnt import tr_cnt
from .trr import trr
from .trr_dist import trr_dist
from .valve import valve
from .width_reductor import width_reductor
from .repack import repack
from .priority_mux_valve import priority_mux_valve
from .max_gears import max2
from .unary import unary

# hier. blocks
from .reduce2 import reduce2

__all__ = [
    'rng', 'iceil', 'priority_mux', 'qcnt', 'sdp', 'chop', 'trr', 'replicate',
    'trr_dist', 'clip', 'din_cat', 'take', 'release_after_eot', 'reverse',
    'valve', 'form_sub_cfg', 'reduce2', 'width_reductor', 'accumulator',
    'tr_cnt', 'repack', 'priority_mux_valve', 'max2', 'qlen_cnt', 'unary'
]



import array
import asyncio
import glob
import itertools
import logging
import math
import os
import socket
import time
from math import ceil
from subprocess import DEVNULL, Popen

import jinja2

from pygears import GearDone, registry
from pygears.definitions import ROOT_DIR
from pygears.sim import clk, sim_log
from pygears.sim.modules.cosim_base import CosimBase, CosimNoData
from pygears.svgen import svgen
from pygears.svgen.util import svgen_typedef
from pygears.typing import Uint
from pygears.typing_common.codec import code, decode
from pygears.util.fileio import save_file


class CosimulatorStartError(Exception):
    pass


async def drive_reset(duration):
    simsoc = registry('SimConfig')['SimSocket']
    await clk()
    data = simsoc.send_req(duration | (1 << 31), Uint[4])
    for i in range(duration):
        await clk()

    return data


def u32_repr_gen(data, dtype):
    yield int(dtype)
    for i in range(ceil(int(dtype) / 32)):
        yield data & 0xffffffff
        data >>= 32


def u32_repr(data, dtype):
    return array.array('I', u32_repr_gen(code(dtype, data), dtype))


def u32_bytes_to_int(data):
    arr = array.array('I')
    arr.frombytes(data)
    val = 0
    for val32 in reversed(arr):
        val <<= 32
        val |= val32

    return val


def u32_bytes_decode(data, dtype):
    return decode(dtype, u32_bytes_to_int(data))


j2_templates = ['runsim.j2', 'top.j2']
j2_file_names = ['run_sim.sh', 'top.sv']


def sv_cosim_gen(gear, tcp_port=1234):
    # pygearslib = util.find_spec("pygearslib")
    # if pygearslib is not None:
    #     from pygearslib import sv_src_path
    #     registry('SVGenSystemVerilogPaths').append(sv_src_path)

    outdir = registry('SimArtifactDir')
    if 'SimSocketHooks' in registry('SimConfig'):
        hooks = registry('SimConfig')['SimSocketHooks']
    else:
        hooks = {}

    srcdir = os.path.join(outdir, 'src_gen')
    rtl_node = svgen(gear, outdir=srcdir)
    sv_node = registry('SVGenMap')[rtl_node]

    port_map = {
        port.basename: port.basename
        for port in itertools.chain(rtl_node.in_ports, rtl_node.out_ports)
    }

    structs = [
        svgen_typedef(port.dtype, f"{port.basename}")
        for port in itertools.chain(rtl_node.in_ports, rtl_node.out_ports)
    ]

    base_addr = os.path.dirname(__file__)
    env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(base_addr),
        trim_blocks=True,
        lstrip_blocks=True)
    env.globals.update(zip=zip, int=int, print=print, issubclass=issubclass)

    context = {
        'intfs': list(sv_node.sv_port_configs()),
        'module_name': sv_node.sv_module_name,
        'dut_name': sv_node.sv_module_name,
        'dti_verif_path': os.path.abspath(
            os.path.join(ROOT_DIR, 'sim', 'dpi')),
        'param_map': sv_node.params,
        'structs': structs,
        'port_map': port_map,
        'out_path': os.path.abspath(outdir),
        'hooks': hooks,
        'rst_mask': "32'h8000_0000",
        'port': tcp_port,
        'activity_timeout': 1000  # in clk cycles
    }

    inc_paths = []
    context['includes'] = []
    for path in registry('SVGenSystemVerilogPaths'):
        inc_paths.append(path)
    inc_paths.append(srcdir)
    inc_paths.append(outdir)

    context['includes'] = [
        os.path.abspath(os.path.join(p, '*.sv')) for p in inc_paths
        if os.path.exists(p)
    ]

    # remove empty wildcard imports
    context['includes'] = [p for p in context['includes'] if glob.glob(p)]

    for templ, tname in zip(j2_templates, j2_file_names):
        res = env.get_template(templ).render(context)
        fname = save_file(tname, context['out_path'], res)
        if os.path.splitext(fname)[1] == '.sh':
            os.chmod(fname, 0o777)


class SimSocketDrv:
    def __init__(self, handler, port):
        self.handler = handler
        self.port = port

    def reset(self):
        pass


class SimSocketInputDrv(SimSocketDrv):
    def close(self):
        self.handler.sendall(b'\x00\x00\x00\x00')
        self.handler.close()
        # del self.handler

    def send(self, data):
        # print(f'Sending {repr(data)} for {self.port.basename} of type {self.port.dtype}')
        pkt = u32_repr(data, self.port.dtype).tobytes()
        self.handler.sendall(pkt)

    def ready(self):
        try:
            self.handler.recv(4)
            return True
        except socket.error:
            return False


class SimSocketOutputDrv(SimSocketDrv):
    def read(self):
        buff_size = math.ceil(int(self.port.dtype) / 8)
        if buff_size < 4:
            buff_size = 4
        if buff_size % 4:
            buff_size += 4 - (buff_size % 4)
        try:
            data = self.handler.recv(buff_size)
            return u32_bytes_decode(data, self.port.dtype)
        except socket.error:
            raise CosimNoData

    def ack(self):
        try:
            self.handler.sendall(b'\x01\x00\x00\x00')
        except socket.error:
            raise GearDone


class SimSocketSynchro:
    def __init__(self, handler):
        self.handler = handler
        self.handler.settimeout(5.0)

    def cycle(self):
        pass

    def forward(self):
        data = None
        while not data:
            try:
                self.handler.sendall(b'\x00\x00\x00\x00')
                data = self.handler.recv(4)
            except socket.timeout:
                import pdb
                pdb.set_trace()
            except socket.error:
                raise GearDone

    back = forward

    def sendall(self, pkt):
        self.handler.sendall(pkt)

    def recv(self, buff_size):
        return self.handler.recv(buff_size)


class SimSocket(CosimBase):
    def __init__(self,
                 gear,
                 rebuild=True,
                 run=False,
                 batch=True,
                 tcp_port=1234,
                 **kwds):
        super().__init__(gear)

        self.rebuild = rebuild

        # Create a TCP/IP socket
        self.run_cosim = run
        kwds['batch'] = batch
        self.kwds = kwds
        self.sock = None

        self.server_address = ('localhost', tcp_port)
        self.handlers = {}

        registry('SimConfig')['SimSocket'] = self

    def finish(self):
        sim_log().debug(f'Closing socket server')
        super().finish()

    def cleanup(self):
        if self.sock:
            sim_log().info(f'Done. Closing the socket...')
            time.sleep(3)
            self.sock.close()
            time.sleep(1)

            if self.cosim_pid is not None:
                self.cosim_pid.terminate()

        super().cleanup()

    def send_req(self, req, dtype):
        # print('SimSocket sending request...')
        data = None

        # Send request
        pkt = req.to_bytes(4, byteorder='little')
        self.handlers[self.SYNCHRO_HANDLE_NAME].sendall(b'\x01\x00\x00\x00' +
                                                        pkt)

        # Get random data
        while data is None:
            try:
                buff_size = math.ceil(int(dtype) / 8)
                if buff_size < 4:
                    buff_size = 4
                if buff_size % 4:
                    buff_size += 4 - (buff_size % 4)
                data = self.handlers[self.SYNCHRO_HANDLE_NAME].recv(buff_size)
            except socket.error:
                sim_log().error(f'socket error on {self.SYNCHRO_HANDLE_NAME}')
                raise socket.error

        data = u32_bytes_decode(data, dtype)
        return data

    def setup(self):
        super().setup()

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # Bind the socket to the port
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        self.sock.bind(self.server_address)

        # Listen for incoming connections
        self.sock.listen(len(self.gear.in_ports) + len(self.gear.out_ports))

        if self.rebuild:
            sv_cosim_gen(self.gear, self.server_address[1])

        self.cosim_pid = None
        if self.run_cosim:

            self.sock.settimeout(1)

            outdir = registry('SimArtifactDir')
            args = ' '.join(f'-{k} {v if not isinstance(v, bool) else ""}'
                            for k, v in self.kwds.items()
                            if not isinstance(v, bool) or v)
            if 'seed' in self.kwds:
                sim_log().warning(
                    'Separately set seed for cosimulator. Ignoring SimRandSeed.'
                )
            else:
                args += f' -seed {registry("SimRandSeed")}'
            if sim_log().isEnabledFor(logging.DEBUG):
                stdout = None
            else:
                stdout = DEVNULL

            sim_log().info(f'Running cosimulator with: {args}')
            self.cosim_pid = Popen(
                [f'./run_sim.sh'] + args.split(' '),
                stdout=stdout,
                stderr=stdout,
                cwd=outdir)
            time.sleep(0.1)
            ret = self.cosim_pid.poll()
            if ret is not None:
                sim_log().error(f"Cosimulator error: {ret}")
                raise CosimulatorStartError
            else:
                sim_log().info(f"Cosimulator started")

        self.loop = asyncio.get_event_loop()

        total_conn_num = len(self.gear.argnames) + len(self.gear.outnames) + 1

        sim_log().info(f'Waiting on {self.sock.getsockname()}')
        while len(self.handlers) != total_conn_num:
            if self.cosim_pid:
                ret = None
                while ret is None:
                    try:
                        conn, addr = self.sock.accept()
                        break
                    except socket.timeout:
                        ret = self.cosim_pid.poll()
                        if ret is not None:
                            sim_log().error(f"Cosimulator error: {ret}")
                            raise Exception
            else:
                sim_log().debug("Wait for connection")
                self.sock.listen(1)
                conn, addr = self.sock.accept()

            msg = conn.recv(1024)
            port_name = msg.decode()

            if port_name == self.SYNCHRO_HANDLE_NAME:
                self.handlers[self.SYNCHRO_HANDLE_NAME] = SimSocketSynchro(
                    conn)
                # conn.setblocking(True)
            else:
                for p in self.gear.in_ports:
                    if p.basename == port_name:
                        self.handlers[port_name] = SimSocketInputDrv(conn, p)
                        break
                for p in self.gear.out_ports:
                    if p.basename == port_name:
                        self.handlers[port_name] = SimSocketOutputDrv(conn, p)
                        break
                conn.setblocking(False)

            sim_log().debug(f"Connection received for {port_name}")



from .base import EnumerableGenericMeta, type_str, type_repr
from .base import TemplatedTypeUnspecified
from .base import class_and_instance_method


class TupleMeta(EnumerableGenericMeta):
    def __new__(cls, name, bases, namespace, args=[]):
        cls = super().__new__(cls, name, bases, namespace, args)

        args = cls.args
        if not args:
            # Generic parameter values have not been supplied
            return cls
        else:
            cls.args = args
            return cls

    def without(self, *index):
        return Tuple[{
            k: v
            for k, v in zip(self.fields, self.args) if k not in index
        }]

    def __add__(self, other):
        return Tuple[{
            **{k: v
               for k, v in zip(self.fields, self.args)},
            **{k: v
               for k, v in zip(other.fields, other.args)}
        }]

    def __repr__(self):
        if not self.args or not hasattr(self, '__parameters__'):
            return super().__repr__()
        else:
            return 'Tuple[{%s}]' % ', '.join([
                f'{repr(f)}: {type_repr(a)}'
                for f, a in zip(self.fields, self.args)
            ])

    # def index_norm(self, index):
    #     if not isinstance(index, tuple):
    #         index = (index, )

    #     index = list(index)
    #     for i, ind in enumerate(index):
    #         if isinstance(ind, str):
    #             try:
    #                 index[i] = self.fields.index(ind)
    #             except ValueError as e:
    #                 raise KeyError(f'Field "{ind}" not in Tuple')

    #     return super().index_norm(tuple(index))

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def __getitem__(self, index):
        if not self.is_specified():
            return super().__getitem__(index)

        index = self.index_norm(index)

        if (len(index) == 1) and (not isinstance(index[0], slice)):
            return self.__args__[index[0]]
        else:
            subtypes = []
            for i in index:
                subt = self.__args__[i]
                subtypes.extend(subt if isinstance(i, slice) else [subt])

            return Tuple[tuple(subtypes)]

    def __str__(self):
        return '(%s)' % ', '.join([type_str(a) for a in self.args])


class Tuple(tuple, metaclass=TupleMeta):
    # def __new__(self, val: tuple):
    def __new__(cls, val):
        if not cls.is_specified():
            raise TemplatedTypeUnspecified

        if type(val) == cls:
            return val

        if isinstance(val, dict):
            tpl_val = tuple(t(val[f]) for t, f in zip(cls, cls.fields))
        else:
            tpl_val = tuple(t(v) for t, v in zip(cls, val))

        return super(Tuple, cls).__new__(cls, tpl_val)

    def __getitem__(self, index):
        index = type(self).index_norm(index)

        if (len(index) == 1) and (not isinstance(index[0], slice)):
            return super(Tuple, self).__getitem__(index[0])
        else:
            tout = type(self)[index]
            subtypes = []
            for i in index:
                subt = super(Tuple, self).__getitem__(i)
                subtypes.extend(subt if isinstance(i, slice) else [subt])

            return tout(tuple(subtypes))

    @class_and_instance_method
    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def __int__(self):
        ret = 0

        for d, t in zip(reversed(self), reversed(type(self))):
            ret <<= int(t)
            ret |= int(d)

        return ret

    @classmethod
    def decode(cls, val):
        ret = []
        for t in cls:
            ret.append(t.decode(val))
            val >>= int(t)

        return cls(tuple(ret))



from pygears import gear, GearDone, module
from logging import info
from pygears.sim import sim_assert, sim_log


@gear
async def scoreboard(*din: b't', report) -> None:
    cnt = 0
    match_cnt = 0
    try:
        while (1):
            items = []
            for d in din:
                items.append(await d.get())

            match = all(v == items[0] for v in items)

            report.append({'match': match, 'items': items})
            cnt += 1
            if match:
                match_cnt += 1
            sim_assert(
                match,
                f'mismatch on #{cnt}: {items[0]}, {items[1]}')

    except GearDone as e:
        sim_log().info(f'Number of matches = {match_cnt}/{cnt}')
        raise e



from pygears import registry
from pygears.sim.sim_gear import SimGear, is_simgear_func
from pygears.sim.sim import SimPlugin
from pygears.core.gear import GearPlugin
from pygears.core.hier_node import HierVisitorBase


class SimInstVisitor(HierVisitorBase):
    def __init__(self):
        self.namespace = registry('SimModuleNamespace')
        self.sim_map = registry('SimMap')

    def Gear(self, module):
        sim_cls = module.params.get('sim_cls', None)
        sim_inst = None

        if sim_cls is None:
            sim_cls = self.namespace.get(module.definition, None)

        if sim_cls:
            # print(f"Recognized {module.name} as {sim_cls}")
            sim_inst = sim_cls(module)
        elif is_simgear_func(module.func):
            # print(f"Recognized {module.name} as sim_gear")
            sim_inst = SimGear(module)

        if sim_inst:
            self.sim_map[module] = sim_inst
            return True


def sim_inst(top):
    v = SimInstVisitor()
    v.visit(top)

    return top


class SimInstPlugin(SimPlugin, GearPlugin):
    @classmethod
    def bind(cls):
        cls.registry['SimFlow'].append(sim_inst)
        cls.registry['SimModuleNamespace'] = {}
        cls.registry['SimMap'] = {}
        cls.registry['GearExtraParams']['sim_cls'] = None

    @classmethod
    def reset(cls):
        cls.registry['SimMap'] = {}



from pygears.typing.visitor import TypingVisitorBase
from pygears.typing import bitw


def svgen_visitor(cls):
    def svgen_action(top, conf):
        v = cls()
        v.conf = conf
        v.visit(top)
        return top

    return svgen_action


class SVGenTypeVisitor(TypingVisitorBase):
    struct_str = "typedef struct packed {"
    union_str = "typedef union packed {"

    def __init__(self, name, basic_type='logic'):
        self.struct_array = []
        self.context = name
        self.basic_type = basic_type
        self.depth = -1

    def visit(self, type_, field):
        self.depth += 1
        type_declaration = super().visit(type_, field)
        self.depth -= 1
        return type_declaration

    def visit_int(self, type_, field):
        if (not self.depth):
            self.struct_array.append(
                f'typedef {self.basic_type} signed [{int(type_)-1}:0] {self.context}_t; // {type_}\n'
            )
        return f'{self.basic_type} signed [{int(type_)-1}:0]'

    def visit_bool(self, type_, field):
        if (not self.depth):
            self.struct_array.append(
                f'typedef {self.basic_type} [0:0] {self.context}_t; // {type_}'
            )
        return f'{self.basic_type} [0:0]'

    def visit_uint(self, type_, field):
        if (int(type_) == 0):
            return None
        if (not self.depth):
            self.struct_array.append(
                f'typedef {self.basic_type} [{int(type_)-1}:0] {self.context}_t; // {type_}\n'
            )
        return f'{self.basic_type} [{int(type_)-1}:0]'

    def visit_unit(self, type_, field):
        return None

    def visit_union(self, type_, field):
        struct_fields = []
        max_len = 0
        high_parent_context = self.context
        middle_parent_context = f'{high_parent_context}_data'
        low_parent_context = f'{middle_parent_context}_data'

        for t in (type_.types):
            if (int(t) > max_len):
                max_len = int(t)

        # for t, f in zip(reversed(type_.args), reversed(type_.fields)):
        #     middle_type = Tuple[{f: t, 'dummy': Uint[max_len-int(t)]}]
        #     struct_fields.append(f'{self.struct_str}')
        #     for tp, fd in zip(reversed(middle_type.args), reversed(middle_type.fields)):
        #         self.context = f'{low_parent_context}_{fd}'
        #         type_declaration = self.visit(tp, fd)
        #         if (type_declaration is not None):
        #             struct_fields.append(f'    {type_declaration} {fd}; // {tp}')
        #     struct_fields.append(f'}} {middle_parent_context}_{f}_t;\n')

        for i, t in reversed(list(enumerate(type_.args))):
            field_tmp = f'f{i}'
            struct_fields.append(f'{self.struct_str} // ({t}, u?)')
            self.context = f'{low_parent_context}_{field_tmp}'
            type_declaration = self.visit(t, field_tmp)
            if (int(t) < max_len):
                struct_fields.append(
                    f'    {self.basic_type} [{max_len-int(t)-1}:0] dummy; // u{max_len-int(t)}'
                )
            if (type_declaration is not None):
                struct_fields.append(f'    {type_declaration} {field_tmp}; // {t}')
            struct_fields.append(f'}} {middle_parent_context}_{field_tmp}_t;\n')

        # create union
        struct_fields.append(f'{self.union_str} // {type_}')
        for i, t in reversed(list(enumerate(type_.args))):
            field_tmp = f'f{i}'
            struct_fields.append(
                f'    {middle_parent_context}_{field_tmp}_t {field_tmp}; // ({t}, u?)')
        struct_fields.append(f'}} {middle_parent_context}_t;\n')

        # create struct
        struct_fields.append(f'{self.struct_str} // {type_}')
        struct_fields.append(
            f'    {self.basic_type} [{bitw(len(type_.args)-1)-1}:0] ctrl; // u{bitw(len(type_.args)-1)}'
        )
        struct_fields.append(f'    {middle_parent_context}_t data; // {type_}')
        struct_fields.append(f'}} {high_parent_context}_t;\n')

        self.struct_array.append('\n'.join(struct_fields))

        self.context = high_parent_context

        return f'{high_parent_context}_t'

    def visit_queue(self, type_, field):
        struct_fields = []
        parent_context = self.context
        struct_fields.append(f'{self.struct_str} // {type_}')
        struct_fields.append(
            f'    {self.basic_type} [{type_.args[1]-1}:0] eot; // u{type_.args[1]}'
        )
        self.context = f'{parent_context}_data'
        type_declaration = self.visit(type_.args[0], type_.fields[0])

        if (type_declaration is not None):
            struct_fields.append(
                f'    {type_declaration} data; // {type_.args[0]}')

        struct_fields.append(f'}} {parent_context}_t;\n')
        self.struct_array.append('\n'.join(struct_fields))

        self.context = parent_context

        return f'{parent_context}_t'

    def visit_tuple(self, type_, field):
        struct_fields = []
        parent_context = self.context
        struct_fields.append(f'{self.struct_str} // {type_}')
        for t, f in zip(reversed(type_.args), reversed(type_.fields)):
            self.context = f'{parent_context}_{f}'
            type_declaration = self.visit(t, f)
            if type_declaration:
                struct_fields.append(f'    {type_declaration} {f}; // {t}')

        struct_fields.append(f'}} {parent_context}_t;\n')
        self.struct_array.append('\n'.join(struct_fields))

        self.context = parent_context

        return f'{parent_context}_t'

    def visit_array(self, type_, field):
        if (not self.depth):
            struct_fields = []
            parent_context = self.context
            self.context = f'{parent_context}_data'

        type_declaration = self.visit(type_.args[0], type_.fields[0])
        split_type = type_declaration.split(" ", 1)
        split_type.insert(1, f'[{type_.args[1]-1}:0]')
        type_declaration = ' '.join(split_type)

        if (not self.depth):
            struct_fields.append(
                f'typedef {type_declaration} {parent_context}_t; // {type_}\n')
            self.struct_array.append('\n'.join(struct_fields))
            self.context = parent_context
            return f'{parent_context}_t'
        else:
            return f'{type_declaration}'


def svgen_typedef(dtype, name):
    if isinstance(dtype, str):
        return f'typedef {dtype} {name}_t;'
    elif int(dtype) == 0:
        return f'typedef logic [0:0] {name}_t;'

    vis = SVGenTypeVisitor(name)
    vis.visit(type_=dtype, field=name)
    return '\n'.join(vis.struct_array)



import inspect

from .base import EnumerableGenericMeta, type_str
from .uint import Uint
from .bool import Bool


class QueueMeta(EnumerableGenericMeta):
    def keys(self):
        return list(range(self.lvl + 1))

    def __new__(cls, name, bases, namespace, args=[]):
        if isinstance(args, dict) and (list(args.values())[1] == 0):
            return list(args.values())[0]
        else:
            return super().__new__(cls, name, bases, namespace, args)

    def __getitem__(self, index):
        if not self.is_specified():
            if inspect.isclass(index) and issubclass(
                    index, Queue) and not self.__args__:
                return Queue[index.args[0], index.lvl + 1]
            elif isinstance(
                    index, tuple) and inspect.isclass(index[0]) and issubclass(
                        index[0], Queue) and not self.__args__:
                return Queue[index[0].args[0], index[0].lvl + index[1]]
            elif isinstance(index, tuple) and (index[1] == 0):
                return index[0]
            else:
                return super().__getitem__(index)

        index = self.index_norm(index)

        lvl = 0
        data_incl = False
        for i in index:
            if isinstance(i, slice):
                if (i.stop == 0) or (i.stop - i.start > self.lvl):
                    raise IndexError
                elif i.start == 0 and i.stop == 1:
                    data_incl = True
                elif i.start == 0 and i.stop > 1:
                    data_incl = True
                    lvl += i.stop - 1
                else:
                    lvl += (i.stop - i.start)
            elif i == 0:
                data_incl = True
            elif i <= self.lvl:
                lvl += 1
            else:
                raise IndexError

        if data_incl and lvl > 0:
            return Queue[self.args[0], lvl]
        elif lvl > 0:
            return Uint[lvl]
        else:
            return self.args[0]

    @property
    def lvl(self):
        return self.args[1]

    @property
    def eot(self):
        return Uint[self.lvl]

    def __str__(self):
        if self.lvl == 1:
            return '[%s]' % type_str(self.args[0])
        else:
            return '[{}]^{}'.format(type_str(self.args[0]), self.lvl)


class Queue(tuple, metaclass=QueueMeta):
    __default__ = [1]
    __parameters__ = ['data', 'eot']

    def __new__(cls, val: tuple):
        if type(val) == cls:
            return val

        queue_tpl = (cls[0](val[0]), *(Bool(v) for v in val[1:]))
        return super(Queue, cls).__new__(cls, queue_tpl)

    @property
    def last(self):
        return self[1:] == ((1 << type(self).lvl) - 1)

    @property
    def eot(self):
        return self[1:]

    @property
    def lvl(self):
        return len(self) - 1

    @property
    def data(self):
        return self[0]

    def __getitem__(self, index):
        index = type(self).index_norm(index)

        lvl = 0
        data_incl = False

        dout = []
        outtype = type(self)[index]
        for i in index:
            if isinstance(i, slice):
                if (i.stop == 0) or (i.stop - i.start > type(self).lvl):
                    raise IndexError
                elif i.start == 0 and i.stop == 1:
                    data_incl = True
                elif i.start == 0 and i.stop > 1:
                    lvl += i.stop - 1
                    data_incl = True
                else:
                    lvl += (i.stop - i.start)

                dout.extend(super().__getitem__(i))
            elif i == 0:
                data_incl = True
                dout.append(super().__getitem__(i))
            elif i <= type(self).lvl:
                lvl += 1
                dout.append(super().__getitem__(i))
            else:
                raise IndexError

        if lvl > 0:
            if data_incl:
                return outtype(tuple(dout))
            else:
                eot = 0
                for d in reversed(dout):
                    eot <<= 1
                    eot |= d

                return outtype(eot)
        else:
            return type(self)[0](dout[0])

    @classmethod
    def decode(cls, val):
        data = [cls[0].decode(val)]

        eot_mask = 1 << int(cls[0])
        for i in range(cls.lvl):
            data.append(Bool(val & eot_mask))
            eot_mask <<= 1

        return cls(tuple(data))



import importlib
import os
import re
import sys


class PluginBase:
    subclasses = []
    registry = {}

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.subclasses.append(cls)
        cls.bind()

    @classmethod
    def clear(cls):
        cls.registry.clear()
        for subc in cls.subclasses:
            subc.bind()

    @classmethod
    def purge(cls):
        cls.subclasses.clear()
        cls.registry.clear()

    @classmethod
    def bind(cls):
        pass

    @classmethod
    def reset(cls):
        pass


def registry(key):
    return PluginBase.registry[key]


def bind(key, val):
    PluginBase.registry[key] = val


def clear():
    PluginBase.clear()


def load_plugin_folder(path):
    plugin_parent_dir, plugin_dir = os.path.split(path)
    sys.path.insert(0, plugin_parent_dir)
    pysearchre = re.compile('.py$', re.IGNORECASE)
    pluginfiles = filter(pysearchre.search, os.listdir(path))
    plugins = map(lambda fp: '.' + os.path.splitext(fp)[0], pluginfiles)
    # import parent module / namespace
    importlib.import_module(plugin_dir)
    modules = []
    for plugin in plugins:
        if not plugin.endswith('__'):
            modules.append(importlib.import_module(plugin, package=plugin_dir))

    return modules



from pygears.core.gear import gear
from pygears.typing import Union
from pygears.common.ccat import ccat


def demux_type(dtypes, ctrl_out):
    if (ctrl_out):
        return (dtypes[1], ) + tuple(t for t in dtypes.types)
    else:
        return tuple(t for t in dtypes.types)


@gear
async def demux(din: Union, *, ctrl_out=False) -> b'demux_type(din, ctrl_out)':
    async with din as item:
        dout = [None] * len(din.dtype.types)

        dout[item[1]] = item.data

        if ctrl_out:
            dout = [item[1]] + dout

        yield tuple(dout)


@gear
def demux_by(ctrl, din, *, fcat=ccat, out_num=None):
    if out_num is None:
        out_num = 2**int(ctrl.dtype)

    return fcat(din, ctrl) \
        | Union[(din.dtype, )*out_num] \
        | demux


@gear
def demux_zip(din: Union) -> b'demux_type(din, True)':
    pass



from pygears import gear
from pygears.sim import sim_assert
from pygears.sim.modules import delay_mon, drv, mon, scoreboard
from pygears.sim.utils import SimDelay
from pygears.common import decoupler


@gear
async def check(din, *, ref):
    try:
        items = []
        while (1):
            items.append(await din.get())
    finally:
        # print(f"Here: {items}")
        # print(f"{ref}")
        # print(f"{items == ref}")
        sim_assert(items == ref, f'mismatch. Got: {items}, expected: {ref}')


def tlm_verif(*seq, f, ref):
    res_tlm = tuple(s | drv for s in seq) \
        | f \
        | mon

    ref_tlm = seq | ref

    report = []
    scoreboard(res_tlm, ref_tlm, report=report)

    return report


def verif(*stim, f, ref, delays=None):
    '''Using ref. model'''
    # if delays is None:
    #     delays = [SimDelay(0, 0)] * (len(seq) + 1)
    # else:
    #     assert len(seq) + 1 == len(delays), print(
    #         'Not enough delays specified')

    # stim = tuple(s | drv(delay=delays[i]) for i, s in enumerate(seq))

    res_tlm = stim | f
    ref_tlm = stim | ref

    if not isinstance(res_tlm, tuple):
        res_tlm = (res_tlm, )
        ref_tlm = (ref_tlm, )

    report = [[] for _ in range(len(res_tlm))]

    if delays is None:
        delays = (None, ) * len(res_tlm)

    assert len(ref_tlm) == len(res_tlm)
    assert len(delays) == len(res_tlm)

    for r, res_intf, ref_intf, d in zip(report, res_tlm, ref_tlm, delays):
        if d is not None:
            res_intf = res_intf | d

        res_intf = res_intf | decoupler(depth=0)
        ref_intf = ref_intf | decoupler(depth=0)

        scoreboard(res_intf, ref_intf, report=r)

    return report


def directed(*stim, f, ref):
    '''Directed test, ref is a list of expected results'''
    res = stim | f
    if isinstance(res, tuple):
        for i, r in enumerate(res):
            r | mon | check(ref=ref[i])
    else:
        res | mon | check(ref=ref)


def directed_on_the_fly(*stim, f, refs, delays=None):
    '''Directed test, but checking done on-the-fly (from generators)'''
    res_tlm = stim | f

    if not isinstance(res_tlm, tuple):
        res_tlm = (res_tlm, )

    if delays is None:
        delays = (None, ) * len(res_tlm)

    report = [[] for _ in range(len(res_tlm))]
    for r, res_intf, ref, d in zip(report, res_tlm, refs, delays):
        if d is not None:
            res_intf = res_intf | d

        res_intf = res_intf | decoupler(depth=0)
        # ref = ref | decoupler(depth=0)

        scoreboard(res_intf, ref, report=r)

    return report



from pygears import module
from pygears.core.gear import gear
from pygears.core.log import gear_log
from pygears.svgen.inst import SVGenInstPlugin
from pygears.svgen.svmod import SVModuleGen
from pygears.typing import Queue


@gear
async def trr(*din: Queue['t_data']) -> b'Queue[t_data, 2]':
    for i, d in enumerate(din):
        val = din[0].dtype((0, 0))
        while (val.eot == 0):
            async with d as val:
                dout = module().tout((val.data, val.eot, (i == len(din) - 1)))
                gear_log().debug(f'Trr yielding {dout}')
                yield dout


class SVGenTrr(SVModuleGen):
    @property
    def is_generated(self):
        return True

    def get_module(self, template_env):
        context = {
            'module_name': self.sv_module_name,
            'intfs': list(self.sv_port_configs())
        }
        return template_env.render_local(__file__, "trr.j2", context)


class SVGenTrrPlugin(SVGenInstPlugin):
    @classmethod
    def bind(cls):
        cls.registry['SVGenModuleNamespace'][trr] = SVGenTrr



from pygears import gear
from pygears.common import ccat


class Identity(dict):
    def __missing__(self, key):
        return key


@gear
def form_sub_cfg_impl(cfg, *, sub_cfg_t, name_map=Identity()):
    sub_cfg_fields = (cfg[name_map[f] if f in name_map else f]
                      for f in sub_cfg_t.fields)
    return ccat(*sub_cfg_fields)


def form_sub_cfg(cfg, sub_cfg_t, name='form_sub_cfg', name_map=Identity()):
    return cfg | form_sub_cfg_impl(
        sub_cfg_t=sub_cfg_t, name_map=name_map, name=name)
    # name_map = Identity(f2='polje2')
    # new_type = rename(bla.dtype, Identity(f2='polje2'))
    # new_type = bla.dtype.base[{name_map[bla.dtype.fields[k]]: v for k, v in bla.dtype.items()}]
    # print(repr(new_type))



from pygears.registry import PluginBase

from .bool import Bool
from .queue import Queue
from .array import Array
from .base import TemplateArgumentsError, typeof, Any
from .tuple import Tuple
from .uint import Int, Uint, Integer
from .unit import Unit
from .union import Union
from .bitw import bitw, ceil_pow2
from .tlm import TLM

__all__ = [
    'Bool', 'Queue', 'TemplateArgumentsError', 'Tuple', 'Int', 'Uint', 'Unit',
    'Union', 'Array', 'bitw', 'typeof', 'Any', 'TLM', 'ceil_pow2'
]


class TypingNamespacePlugin(PluginBase):
    @classmethod
    def bind(cls):
        cls.registry['TypeArithNamespace'] = {}


class CoreTypesPlugin(TypingNamespacePlugin):
    @classmethod
    def bind(cls):
        cls.registry['TypeArithNamespace']['Union'] = Union
        cls.registry['TypeArithNamespace']['Tuple'] = Tuple
        cls.registry['TypeArithNamespace']['Uint'] = Uint
        cls.registry['TypeArithNamespace']['Int'] = Int
        cls.registry['TypeArithNamespace']['Integer'] = Integer
        cls.registry['TypeArithNamespace']['Unit'] = Unit
        cls.registry['TypeArithNamespace']['Bool'] = Bool
        cls.registry['TypeArithNamespace']['Queue'] = Queue
        cls.registry['TypeArithNamespace']['Array'] = Array
        cls.registry['TypeArithNamespace']['bitw'] = bitw
        cls.registry['TypeArithNamespace']['ceil_pow2'] = ceil_pow2
        cls.registry['TypeArithNamespace']['typeof'] = typeof
        cls.registry['TypeArithNamespace']['Any'] = Any
        cls.registry['TypeArithNamespace']['TLM'] = TLM



import asyncio

from pygears import registry, GearDone
from pygears.core.port import InPort, OutPort
from pygears.registry import PluginBase
from pygears.core.sim_event import SimEvent
from pygears.typing.base import TypingMeta


def operator_func_from_namespace(cls, name):
    def wrapper(self, *args, **kwargs):
        if name not in registry('IntfOperNamespace'):
            raise Exception(f'Operator {name} is not supported.')

        operator_func = registry('IntfOperNamespace')[name]
        return operator_func(self, *args, **kwargs)

    return wrapper


def operator_methods_gen(cls):
    for name in cls.OPERATOR_SUPPORT:
        setattr(cls, name, operator_func_from_namespace(cls, name))
    return cls


def _get_consumer_tree_rec(intf, consumers):
    for port in intf.consumers:
        cons_intf = port.consumer
        if (port.gear in registry('SimMap')) and (isinstance(port, InPort)):
            # if not cons_intf.consumers:
            consumers.append(port)
        else:
            _get_consumer_tree_rec(cons_intf, consumers)


def get_consumer_tree(intf):
    consumers = []
    _get_consumer_tree_rec(intf, consumers)
    return consumers


@operator_methods_gen
class Intf:
    OPERATOR_SUPPORT = [
        '__getitem__', '__neg__', '__add__', '__sub__', '__mul__',
        '__div__', '__floordiv__', '__mod__', '__invert__', '__rshift__',
        '__lt__'
    ]

    def __init__(self, dtype):
        self.consumers = []
        self._end_consumers = None
        self.dtype = dtype
        self.producer = None
        self._in_queue = None
        self._out_queues = []
        self._done = False
        self._data = None

        self.events = {
            'put': SimEvent(),
            'put_out': SimEvent(),
            'ack': SimEvent(),
            'ack_in': SimEvent(),
            'pull_start': SimEvent(),
            'pull_done': SimEvent(),
            'cancel': SimEvent()
        }

    def __ior__(self, iout):
        return iout.__matmul__(self)

    def __or__(self, other):
        if not isinstance(other, (str, TypingMeta)):
            return other.__ror__(self)

        operator_func = registry('IntfOperNamespace')['__or__']
        return operator_func(self, other)

    def __matmul__(self, iout):
        self.producer.consumer = iout
        iout.producer = self.producer

        return iout

    def source(self, port):
        self.producer = port
        port.consumer = self

    def disconnect(self, port):
        if port in self.consumers:
            self.consumers.remove(port)
            port.producer = None
        elif port == self.producer:
            port.consumer = None
            self.producer = None

    def connect(self, port):
        self.consumers.append(port)
        port.producer = self

    @property
    def end_consumers(self):
        if self._end_consumers is None:
            self._end_consumers = get_consumer_tree(self)

        return self._end_consumers

    @property
    def in_queue(self):
        if self._in_queue is None:
            if self.producer is not None:
                self._in_queue = self.producer.get_queue()

        return self._in_queue

    def get_consumer_queue(self, port):
        for pout in self.consumers:
            if pout.gear in registry('SimMap') and (isinstance(pout, OutPort)):
                out_queues = self.out_queues
                try:
                    i = self.end_consumers.index(port)
                except Exception as e:
                    print(
                        f'Port {port.gear.name}.{port.basename} not in end consumer list of {self.consumers[0].gear.name}.{self.consumers[0].basename}'
                    )
                    raise e
                return out_queues[i]
        else:
            if self.producer:
                return self.producer.get_queue(port)
            else:
                raise Exception(
                    f'Interface path does not end with a simulation gear at {pout.gear.name}.{pout.basename}'
                )

    @property
    def out_queues(self):
        if self._out_queues:
            return self._out_queues

        # if len(self.consumers) == 1 and self.in_queue:
        # if self.producer is not None:
        #     return [self.in_queue]
        # else:
        self._out_queues = [
            asyncio.Queue(maxsize=1) for _ in self.end_consumers
        ]

        for i, q in enumerate(self._out_queues):
            q.intf = self
            q.index = i

        return self._out_queues

    def put_nb(self, val):
        if any(registry('SimMap')[c.gear].done for c in self.end_consumers):
            raise GearDone

        e = self.events['put']

        if e:
            e(self, val)

        for q, c in zip(self.out_queues, self.end_consumers):
            if e:
                e(c.consumer, val)

            q.put_nowait(val)

    async def ready(self):
        if not self.ready_nb():
            for q, c in zip(self.out_queues, self.end_consumers):
                registry('CurrentModule').phase = 'back'
                await q.join()

    def ready_nb(self):
        return all(not q._unfinished_tasks for q in self.out_queues)

    async def put(self, val):
        self.put_nb(val)
        await self.ready()

    def empty(self):
        if self._data is not None:
            return False
        else:
            return self.in_queue.empty()

    def finish(self):
        self._done = True
        for q, c in zip(self.out_queues, self.end_consumers):
            c.finish()
            for task in q._getters:
                self.events['cancel'](self, c)
                task.cancel()

    def done(self):
        return self._done

    def pull_nb(self):
        if self._done:
            raise GearDone

        if self._data is None:
            self._data = self.in_queue.get_nowait()

        return self._data

    def get_nb(self):
        val = self.pull_nb()
        self.ack()
        return val

    async def pull(self):
        e = self.events['pull_start']
        if e:
            e(self)

        if self._done:
            raise GearDone

        if self._data is None:
            self._data = await self.in_queue.get()

        e = self.events['pull_done']
        if e:
            e(self)

        return self._data

    def ack(self):
        e = self.events['ack']
        if e:
            e(self)

        ret = self.in_queue.task_done()
        if self.in_queue.intf.ready_nb():
            e(self.in_queue.intf)

        self._data = None
        return ret

    async def get(self):
        val = await self.pull()
        self.ack()
        return val

    async def __aenter__(self):
        return await self.pull()

    async def __aexit__(self, exception_type, exception_value, traceback):
        if exception_type is None:
            self.ack()

    def __hash__(self):
        return id(self)


class IntfOperPlugin(PluginBase):
    @classmethod
    def bind(cls):
        cls.registry['IntfOperNamespace'] = {}



# from pygears.registry import load_plugin_folder
# import os
# load_plugin_folder(os.path.join(os.path.dirname(__file__), 'modules'))

from .sim import sim, artifacts_dir, sim_assert, timestep, clk, delta, sim_log, sim_phase

from . import inst

from .sim import SimPlugin

__all__ = [
    'sim', 'artifacts_dir', 'sim_assert', 'clk', 'delta', 'timestep',
    'sim_log', 'sim_phase'
]



import itertools

from pygears.common.sieve import sieve
from pygears.svgen.inst import SVGenInstPlugin
from pygears.svgen.svmod import SVModuleGen
from functools import partial
from pygears.svgen.svgen import SVGenPlugin
from pygears.svgen.util import svgen_visitor
from pygears.core.hier_node import HierVisitorBase
from pygears.svgen.inst import svgen_inst
from pygears.rtl.gear import RTLGearHierVisitor, is_gear_instance


def index_to_sv_slice(dtype, index):
    subtype = dtype[index]

    if isinstance(index, slice):
        index = index.start

    if index is None or index == 0:
        low_pos = 0
    else:
        low_pos = int(dtype[:index])

    high_pos = low_pos + int(subtype) - 1

    return f'{high_pos}:{low_pos}'


class SVGenSieve(SVModuleGen):
    @property
    def is_generated(self):
        return True

    def get_module(self, template_env):
        def get_stages():
            for s in itertools.chain(self.node.pre_sieves, [self.node]):
                indexes = s.params['index']
                if not isinstance(indexes, tuple):
                    indexes = (indexes, )

                dtype = s.in_ports[0].dtype
                out_type = s.out_ports[0].dtype
                slices = list(
                    map(
                        partial(index_to_sv_slice, dtype),
                        filter(lambda i: int(dtype[i]) > 0, indexes)))
                yield slices, out_type

        stages = list(get_stages())
        # If any of the sieves has shrunk data to 0 width, there is nothing to
        # do
        if any(i[0] == [] for i in stages):
            stages = []

        context = {
            'stages': stages,
            'module_name': self.sv_module_name,
            'intfs': list(self.sv_port_configs())
        }

        return template_env.render_local(__file__, "sieve.j2", context)


@svgen_visitor
class RemoveEqualReprSieveVisitor(RTLGearHierVisitor):
    def sieve(self, node):
        pout = node.out_ports[0]
        pin = node.in_ports[0]

        if pin.dtype == pout.dtype:
            node.bypass()


@svgen_visitor
class CollapseSievesVisitor(RTLGearHierVisitor):
    def sieve(self, node):
        if not hasattr(node, 'pre_sieves'):
            node.pre_sieves = []

        sieve_cons = [
            p for p in node.consumers if is_gear_instance(p.node, sieve)
        ]
        pin = node.in_ports[0]
        pout = node.out_ports[0]
        iin = pin.producer
        iout = pout.consumer

        if sieve_cons:
            # There is a Sieve connected to this Sieve, hence we can combine
            # two of them into a single SV module

            # Connect the consumers of this Sieve, which are Sieves themselves,
            # to this Sieve's predecessor
            for cons_pin in iout.consumers.copy():
                consumer = cons_pin.node
                if is_gear_instance(consumer, sieve):
                    # print(f'Merging {node.name} to {consumer.name}')
                    # print(consumer.params['index'])
                    # If the consumer is a Sieve, just register this Sieve with
                    # it, and short circuit this one
                    consumer.pre_sieves = node.pre_sieves + [node]
                    iout.disconnect(cons_pin)
                    iin.connect(cons_pin)

            # print(f'Remaining conusmer: {[p.node.name for p in node.consumers]}')

            if not node.consumers:
                # Finally, if ther are no consumers left for this sieve remove
                # this Sieve completely (with all it's connections) from the
                # SVGen tree
                node.remove()
                iout.remove()


class SVGenSievePlugin(SVGenInstPlugin, SVGenPlugin):
    @classmethod
    def bind(cls):
        cls.registry['SVGenModuleNamespace'][sieve] = SVGenSieve
        cls.registry['SVGenFlow'].insert(
            cls.registry['SVGenFlow'].index(svgen_inst),
            CollapseSievesVisitor)
        # cls.registry['SVGenFlow'].insert(
        #     cls.registry['SVGenFlow'].index(CollapseSievesVisitor),
        #     RemoveEqualReprSieveVisitor)



from pygears.svgen.svmod import SVModuleGen
from pygears.svgen.inst import SVGenInstPlugin
from pygears import gear
from pygears import module, alternative
from pygears.sim import clk
from pygears.typing import Queue, Union


# TODO: why is b' necessary in return expression?
@gear(enablement=b'not all(typeof(d, Queue) for d in din)')
async def priority_mux(*din) -> b'Union[din]':
    for i, d in enumerate(din):
        if not d.empty():
            async with d as item:
                yield module().tout((item, i))

            break

    await clk()


def prio_mux_queue_type(dtypes):
    utypes = (dtypes[0][0], ) * len(dtypes)
    return Queue[Union[utypes], dtypes[0].lvl]


@alternative(priority_mux)
@gear(enablement=b'all(typeof(d, Queue) for d in din)')
async def priority_mux_queue(*din) -> b'prio_mux_queue_type(din)':
    pass


class SVGenPriorityMux(SVModuleGen):
    @property
    def is_generated(self):
        return True

    def get_module(self, template_env):
        context = {
            'module_name':
            self.sv_module_name,
            'intfs':
            list(self.sv_port_configs()),
            'is_queue':
            all(
                map(lambda i: issubclass(i['type'], Queue),
                    self.sv_port_configs()))
        }
        return template_env.render_local(__file__, "priority_mux.j2", context)


class SVGenPriorityMuxPlugin(SVGenInstPlugin):
    @classmethod
    def bind(cls):
        cls.registry['SVGenModuleNamespace'][priority_mux] = SVGenPriorityMux
        cls.registry['SVGenModuleNamespace'][
            priority_mux_queue] = SVGenPriorityMux



from pygears import module, gear
from pygears.typing.base import TypingMeta, typeof
from pygears.core.intf import IntfOperPlugin
from pygears.typing_common.codec import code, decode
from pygears.typing import Int, Tuple, Queue, Uint, Union


@gear
async def cast(din, *, cast_type) -> b'cast(din, cast_type)':
    async with din as d:
        if typeof(cast_type,
                  Int) and (not cast_type.is_specified()) and typeof(
                      din.dtype, (Uint, Int)):
            dout = module().tout(d)
        elif typeof(cast_type, Tuple) and typeof(
                din.dtype, Queue) and not cast_type.is_specified():
            dout = module().tout((d[0], d[1:]))
        elif (typeof(cast_type, Union) and typeof(din.dtype, Tuple)
              and len(din.dtype) == 2 and not cast_type.is_specified()):
            pass
        else:
            dout = decode(module().tout, code(din.dtype, d))

        yield dout


def pipe(self, other):
    if self.producer is not None:
        name = f'cast_{self.producer.basename}'
    else:
        name = 'cast'

    return cast(self, cast_type=other, name=name)


class PipeIntfOperPlugin(IntfOperPlugin):
    @classmethod
    def bind(cls):
        cls.registry['IntfOperNamespace']['__or__'] = pipe



import random

from pygears.sim import clk


class SimDelay:
    def __init__(self, low, high):
        self.low = low
        self.high = high

    @property
    async def delay(self):
        for i in range(random.randint(self.low, self.high)):
            await clk()



from pygears import gear
from pygears.common import ccat

# @gear
# def repack(cfg, *, t, name_map={}, val_map={}):

#     mapped_names = [name_map.get(f, f) for f in dtype.fields]
#     map_dict = {n: val.get(n, val_map.get(n, None)) for n in mapped_names}

#     return dtype(map_dict)

#     fields = (cfg[name_map.get(f, f)] for f in t.fields)
#     return ccat(*fields)


@gear
def repack(cfg, *, t, name_map={}, val_map={}):
    fields = []
    for f_name in t.fields:
        if f_name in val_map:
            f = val_map[f_name]
        else:
            f_name = name_map.get(f_name, f_name)
            f = cfg[f_name]

        fields.append(f)

    return ccat(*fields) | t



from pygears import module
from pygears.sim import delta
from pygears.typing import Queue, Tuple, typeof
from pygears.core.gear import gear, alternative
from .ccat import ccat
from .permute import permuted_apply
from functools import reduce
from pygears.util.utils import gather


def lvl_if_queue(t):
    if not issubclass(t, Queue):
        return 0
    else:
        return t.lvl


def data_if_queue(t):
    if not issubclass(t, Queue):
        return t
    else:
        return t[0]


def zip_type(dtypes):
    arg_queue_lvl = list(map(lvl_if_queue, dtypes))
    base_type = Tuple[tuple(map(data_if_queue, dtypes))]

    # If there are no Queues, i.e. max(arg_queue_lvl) == 0, the type below
    # will resolve to just base_type
    return Queue[base_type, max(arg_queue_lvl)]


@gear
async def zip_cat(*din) -> b'zip_type(din)':
    id_max_lvl, max_lvl = max(
        enumerate(din),
        key=lambda p: p[1].dtype.lvl if typeof(p[1].dtype, Queue) else 0)

    async with gather(*din) as dout:
        data = tuple(d.data if typeof(d, Queue) else d for d in dout)
        eots = dout[id_max_lvl].eot

        yield module().tout((data, *eots))


def isort(iterable, key=lambda x: x, reverse=False):
    res = sorted(enumerate(iterable), key=key, reverse=reverse)
    values = tuple(d[1] for d in res)
    indices = tuple(d[0] for d in res)

    return values, indices


@gear(enablement=b'len(din) == 2')
def czip(*din) -> b'zip_type(din)':
    return din | zip_sync(outsync=False) | zip_cat


@alternative(czip)
@gear(enablement=b'len(din) > 2')
def czip_vararg(*din):
    # Sort input interfaces in descending order of their Queue levels, i.e. we
    # want to zip highest Queue levels first in order to synchronize them first
    din_sorted_by_lvl, din_sort_indices = isort(
        din, key=lambda x: lvl_if_queue(x[1].dtype), reverse=True)

    # Zip din's in sorted order using it as a binary operation. This will
    # produce nested Tuple's, hence we cast it to a Queue of single Tuple
    ret_flat_type = zip_type([d.dtype for d in din_sorted_by_lvl])

    def czip_cascade(*din):
        return reduce(czip, din) | ret_flat_type

    return permuted_apply(*din, f=czip_cascade, indices=din_sort_indices)


@gear
def unzip(din, *, dtypes):
    zdata = din[0]
    zlast = din[1:]

    def split():
        for i, d in enumerate(dtypes):
            data = zdata[i]
            if issubclass(d, Queue):
                yield ccat(data, zlast[:d.lvl]) | Queue[data.dtype, d.lvl]
            else:
                yield data

    return tuple(split())


@gear(enablement=b'len(din) == 2')
async def zip_sync(*din, outsync=True) -> b'din':
    lvls = (d.dtype.lvl if typeof(d.dtype, Queue) else 0 for d in din)
    overlap_lvl = min(lvls)

    eot_aligned = (1, 1)

    while(1):

        din_data = tuple(await d.pull() for d in din)

        eot_overlap = [d[:overlap_lvl] for d in din_data]

        eot_aligned = (eot_overlap[0] >= eot_overlap[1], eot_overlap[1] >= eot_overlap[0])

        if all(eot_aligned):
            yield din_data
        else:
            await delta()

            for d, aligned in zip(din, eot_aligned):
                if not aligned:
                    d.ack()


@alternative(zip_sync)
@gear(enablement=b'len(din) > 2')
def zip_sync_vararg(*din):
    return din | czip | unzip(dtypes=[d.dtype for d in din])



from pygears.typing import TypingNamespacePlugin, Queue, Tuple, Union, typeof


def next_pos(type_list, comb, t):
    if len(type_list) == 1:
        yield comb + [t]
    else:
        yield from type_comb_rec(type_list[:-1], comb + [t])


def type_comb_rec(type_list, comb):
    type_ = type_list[-1]
    if typeof(type_, Union):
        for t in type_.types:
            yield from next_pos(type_list, comb, t)
    else:
        yield from next_pos(type_list, comb, type_)


def tuple_type_comb(type_):
    type_list = [t for t in type_]

    for t in type_comb_rec(type_list, []):
        yield Tuple[tuple(reversed(t))]


def queue_type_comb(type_):
    # type_list = [t for t in type_[0]]
    for t in type_[0]:
        yield Queue[t, type_.lvl]

    # for t in type_comb_rec(type_list, []):
    #     yield Queue[tuple(reversed(t)), type_.lvl]


def expand(type_):
    if typeof(type_, Tuple):
        return Union[tuple(tuple_type_comb(type_))]
    elif typeof(type_, Queue):
        if typeof(type_[0], Tuple):
            return Tuple[tuple(queue_type_comb(type_))]
        elif typeof(type_[0], Union):
            utypes = [Queue[t, type_.lvl] for t in type_[0].types]
            return Union[tuple(utypes)]
    else:
        return type_


class TypeExpandPlugin(TypingNamespacePlugin):
    @classmethod
    def bind(cls):
        cls.registry['TypeArithNamespace']['expand'] = expand



from pygears.rtl.node import RTLNode
from pygears.rtl.intf import RTLIntf
from pygears.core.port import InPort
import itertools
from pygears import registry
from pygears.core.hier_node import HierNode
from pygears import registry, PluginBase
from pygears.core.hier_node import HierVisitorBase
import inspect


class RTLGear(RTLNode):
    def __init__(self, gear, parent):
        super().__init__(parent, gear.basename, params=gear.params)
        self.gear = gear


def is_gear_instance(node, definition):
    if isinstance(node, RTLGear):
        return node.gear.definition is definition

    return False


class RTLGearHierVisitor(HierVisitorBase):
    def RTLGear(self, node):
        gear = node.gear
        if hasattr(self, gear.definition.__name__):
            return getattr(self, gear.definition.__name__)(node)


def rtl_from_gear_port(gear_port):
    node_gen = registry('RTLNodeMap').get(gear_port.gear, None)
    rtl_port = None
    if node_gen:
        node = node_gen.node

        if isinstance(gear_port, InPort):
            port_group = node.in_ports
        else:
            port_group = node.out_ports

        rtl_port = port_group[gear_port.index]

    return rtl_port


class RTLGearNodeGen(HierNode):
    def __init__(self, gear, parent):
        super().__init__(parent)
        self.gear = gear
        self.node = RTLGear(gear, getattr(parent, "node", None))

        namespace = registry('SVGenModuleNamespace')

        if 'svgen' not in self.node.params:
            self.node.params['svgen'] = {}

        if 'svgen_cls' not in self.node.params['svgen']:
            svgen_cls = namespace.get(gear.definition, None)

            if svgen_cls is None:
                for base_class in inspect.getmro(gear.__class__):
                    if base_class.__name__ in namespace:
                        svgen_cls = namespace[base_class.__name__]
                        break

            self.node.params['svgen']['svgen_cls'] = svgen_cls

        for p in gear.in_ports:
            self.node.add_in_port(p.basename, p.producer, p.consumer, p.dtype)

        for p in gear.out_ports:
            self.node.add_out_port(p.basename, p.producer, p.consumer, p.dtype)

    def connect(self):
        self.rtl_map = registry('RTLNodeMap')

        for p, gear_p in zip(self.node.in_ports, self.gear.in_ports):
            self.create_intf(p, gear_p, domain=self.node)
            prod_intf = gear_p.producer
            if (self.node.parent is not None and prod_intf is not None
                    and prod_intf.producer is None):
                self.create_unsourced_intf(p, gear_p)

        for p, gear_p in zip(self.node.out_ports, self.gear.out_ports):
            self.create_intf(p, gear_p, domain=self.node.parent)
            gear_intf = gear_p.consumer
            if self.node.parent is not None and gear_intf is not None and not gear_intf.consumers:
                # intf_inst = RTLIntf(
                #     self.node.root(), gear_intf.dtype)

                self.node.root().add_out_port(p.basename, dtype=p.dtype)

                # intf_inst.producer = self.node.root().out_ports[-1]

                self.rtl_map[gear_intf].connect(self.node.root().out_ports[-1])

    def create_unsourced_intf(self, port, gear_port):
        gear_intf = gear_port.producer
        consumers = []
        for cons_port in gear_intf.consumers:
            rtl_port = rtl_from_gear_port(cons_port)
            if rtl_port:
                consumers.append(rtl_port)

        intf_inst = RTLIntf(
            self.node.root(), gear_intf.dtype, consumers=consumers)

        self.node.root().add_in_port(
            port.basename, dtype=intf_inst.dtype, consumer=intf_inst)

        intf_inst.producer = self.node.root().in_ports[-1]

        for cons_port in consumers:
            cons_port.producer = intf_inst

        self.rtl_map[gear_intf] = intf_inst

    def create_intf(self, port, gear_port, domain):
        gear_intf = gear_port.consumer
        if gear_intf is not None:
            consumers = []
            for cons_port in gear_intf.consumers:
                rtl_port = rtl_from_gear_port(cons_port)
                if rtl_port:
                    consumers.append(rtl_port)

            intf_inst = RTLIntf(
                domain, gear_intf.dtype, producer=port, consumers=consumers)

            if hasattr(gear_intf, 'var_name'):
                intf_inst.var_name = gear_intf.var_name

            for cons_port in consumers:
                cons_port.producer = intf_inst

            self.rtl_map[gear_intf] = intf_inst
            port.consumer = intf_inst


class RTLInstPlugin(PluginBase):
    @classmethod
    def bind(cls):
        cls.registry['RTLModuleNamespace'] = {}
        cls.registry['RTLClsMap'] = {}



from pygears import gear, alternative
from pygears.core.gear import GearTypeNotSpecified
from pygears.typing import Uint, Int, Tuple, bitw, Integer
from pygears.typing.base import TypingMeta


def get_int_type(val):
    if val == 0:
        return Uint[1]
    elif val > 0:
        return Uint[bitw(val)]
    else:
        return Int[bitw(val) + 1]


def get_literal_type(val):
    if isinstance(type(val), TypingMeta):
        return type(val)
    elif isinstance(val, int):
        return get_int_type(val)
    elif isinstance(val, tuple):
        dtypes = [get_literal_type(v) for v in val]
        return Tuple[tuple(dtypes)]
    else:
        raise GearTypeNotSpecified(
            f"Value {val} not supported for const module")


@gear(svgen={'svmod_fn': 'sustain.sv'})
async def const(*, val, tout=b'get_literal_type(val)') -> b'tout':
    yield tout(val)


@gear(svgen={'svmod_fn': 'sustain_ping.sv'})
async def const_ping(ping, *, val, tout=b'get_literal_type(val)') -> b'tout':
    async with ping:
        yield tout(val)



from pygears.core.gear import alternative, gear
from pygears.typing import Integer, Int, Uint
from pygears.core.intf import IntfOperPlugin
from pygears.util.hof import oper_reduce


def mod_type(dtypes):
    length = int(dtypes[1])

    return Uint[length]


@gear(svgen={'svmod_fn': 'mod.sv'}, enablement=b'len(din) == 2')
def mod(*din: Integer, din0_signed=b'typeof(din0, Int)',
        din1_signed=b'typeof(din1, Int)') -> b'mod_type(din)':
    pass


@alternative(mod)
@gear
def mod_vararg(*din: Integer, enablement=b'len(din) > 2') -> b'mod_type(din)':
    return oper_reduce(din, mod)


class ModIntfOperPlugin(IntfOperPlugin):
    @classmethod
    def bind(cls):
        cls.registry['IntfOperNamespace']['__mod__'] = mod



import re

from pygears.registry import registry
from pygears.typing.base import Any, GenericMeta, type_repr


class TypeMatchError(Exception):
    pass


def _type_match_rec(t, pat, matches, allow_incomplete):

    # Ignore parameter names in type
    if isinstance(t, str):
        pass
    # Did we reach the parameter name?
    elif isinstance(pat, str):
        if pat in matches:
            # If the parameter name is already bound, check if two deductions
            # are same
            if repr(t) != repr(matches[pat]) and t != Any and matches[pat] != Any:
                raise TypeMatchError(
                    f"Ambiguous match for parameter {pat}: {type_repr(t)} "
                    f"and {type_repr(matches[pat])}")
        else:
            try:
                res = eval(pat, registry('TypeArithNamespace'), matches)
                if repr(t) != repr(res):
                    raise TypeMatchError(
                        f"{type_repr(t)} cannot be matched to {type_repr(res)}"
                    )
            except Exception as e:
                matches[pat] = t
                # if not allow_incomplete:
                # raise TypeMatchError(f"Cannot evaluate {type_repr(pat)}")
    elif (t == Any) or (pat == Any):
        pass
    elif isinstance(t, GenericMeta) and isinstance(
            pat, GenericMeta) and (issubclass(t.base, pat.base)):
        for ta, pa in zip(t.args, pat.args):
            try:
                _type_match_rec(
                    ta, pa, matches, allow_incomplete=allow_incomplete)
            except TypeMatchError as e:
                raise TypeMatchError(
                    f'{str(e)}\n - when matching {repr(t)} to {repr(pat)}')
    elif t == pat:
        pass
    else:
        raise TypeMatchError("{} cannot be matched to {}".format(
            type_repr(t), type_repr(pat)))


def type_match(t, pat, matches={}, allow_incomplete=False):
    matches = dict(matches)
    _type_match_rec(t, pat, matches, allow_incomplete=allow_incomplete)
    return matches



from pygears.svgen.svmod import SVModuleGen
from pygears.svgen.inst import SVGenInstPlugin
from pygears import gear
from pygears.typing import Union


def fill_type(din_t, union_t, field_sel):
    dtypes = []
    for i, t in enumerate(union_t.types):
        if(i == field_sel):
            dtypes.append(din_t)
        else:
            dtypes.append(t)
    return Union[tuple(dtypes)]


@gear
def fill(din,
         union_din: Union,
         *,
         field_sel) -> b'fill_type(din, union_din, field_sel)':
    pass


class SVGenFill(SVModuleGen):
    @property
    def is_generated(self):
        return True

    def get_module(self, template_env):
        context = {
            'module_name': self.sv_module_name,
            'intfs': list(self.sv_port_configs()),
            'field_sel': self.node.params['field_sel']
        }
        return template_env.render_local(__file__, "fill.j2", context)


class SVGenFillPlugin(SVGenInstPlugin):
    @classmethod
    def bind(cls):
        cls.registry['SVGenModuleNamespace'][fill] = SVGenFill



import inspect

from .base import EnumerableGenericMeta


class TLMMeta(EnumerableGenericMeta):
    def __getitem__(self, index):
        if not self.is_specified():
            return super().__getitem__(index)

        index = self.index_norm(index)

        if len(index) != 1:
            raise IndexError

        return self.args[0]


class TLM(metaclass=TLMMeta):
    __parameters__ = ['T']



from pygears.core.gear import alternative, gear
from pygears.typing import Queue, Tuple, typeof
from pygears.common.shred import shred
from pygears.common.ccat import ccat
from pygears.util.utils import quiter_async
from pygears import module


def lvl_if_queue(t):
    if not issubclass(t, Queue):
        return 0
    else:
        return t.lvl


def cart_type(dtypes):
    arg_queue_lvl = [lvl_if_queue(d) for d in dtypes]

    base_type = Tuple[tuple(
        d if lvl == 0 else d[0] for d, lvl in zip(dtypes, arg_queue_lvl))]

    # If there are no Queues, i.e. sum(arg_queue_lvl) == 0, the type below
    # will resolve to just base_type
    return Queue[base_type, sum(arg_queue_lvl)]


@gear(enablement=b'len(din) == 2')
async def cart(*din) -> b'cart_type(din)':
    din_t = [d.dtype for d in din]

    if all(typeof(t, Queue) for t in din_t):
        queue_id, single_id = 1, 0
    elif typeof(din_t[0], Queue):
        queue_id, single_id = 0, 1
    else:
        queue_id, single_id = 1, 0

    async with din[single_id] as single_data:
        if typeof(din_t[single_id], Queue):
            single_eot = single_data.eot
            single_data = single_data.data
        else:
            single_eot = []

        async for queue_data in quiter_async(din[queue_id]):
            out_data = [0, 0]
            out_data[queue_id] = queue_data.data
            out_data[single_id] = single_data

            yield module().tout((tuple(out_data), *queue_data.eot,
                                 *single_eot))


@alternative(cart)
@gear
def cart_vararg(*din, enablement=b'len(din) > 2'):
    ret = cart(din[0], din[1])
    for d in din[2:]:
        ret = cart(ret, d)

    return ret | cart_type([d.dtype for d in din])


# TODO: Lowest eot for each uncart output needs to be shortened to 1 data using
# flattening
@gear
def uncart(din, *, dtypes):
    zdata = din[0]
    zlast = din[1:]

    def split():
        for i, d in enumerate(dtypes):
            data = zdata[i]
            if issubclass(d, Queue):
                yield ccat(data, zlast[:d.lvl]) | Queue[data.dtype, d.lvl]
            else:
                yield data

    return tuple(split())


@gear(enablement=b'len(din) == 2')
async def cart_sync(*din) -> b'din':
    din_t = [d.dtype for d in din]

    queue_id, single_id = (0, 1) if typeof(din_t[0], Queue) else (1, 0)

    async with din[single_id] as single_data:
        async for queue_data in quiter_async(din[queue_id]):
            dout = [0, 0]
            dout[single_id] = single_data
            dout[queue_id] = queue_data
            yield tuple(dout)


@alternative(cart_sync)
@gear
def cart_sync_vararg(*din):
    return din | cart | uncart(dtypes=[d.dtype for d in din])


@gear
def cart_sync_with(sync_in, din, *, balance=None):
    if balance:
        sync_in = sync_in | balance

    din_sync, sync_in_sync = cart_sync(din, sync_in)
    sync_in_sync | shred

    return din_sync



from pygears.svgen.inst import SVGenInstPlugin
from pygears.svgen.svmod import SVModuleGen
from pygears.common.local_rst import local_rst


class SVGenLocalRst(SVModuleGen):
    def get_inst(self, template_env):
        in_intf_name = self.get_in_port_map_intf_name(self.node.in_ports[0])

        return f"""local_rst local_rst_i (
    .clk(clk),
    .rst(rst),
    .local_rst(local_rst),
    .din({in_intf_name})
);
"""


class SVGenAlignPlugin(SVGenInstPlugin):
    @classmethod
    def bind(cls):
        cls.registry['SVGenModuleNamespace'][local_rst] = SVGenLocalRst



import os
import re

from vcd.gtkw import GTKWSave

from pygears import find, registry
from pygears.common.decoupler import decoupler_din
from pygears.core.port import OutPort as GearOutPort
from pygears.rtl.port import InPort, OutPort
from pygears.sim import sim_log
from pygears.sim.extens.graphviz import graph
from pygears.sim.extens.vcd import module_sav
from pygears.sim.modules.sim_socket import SimSocket


def _get_end_consumer_rec(intf, consumers):
    for port in intf.consumers:
        cons_intf = port.consumer

        if isinstance(port, InPort) and (not cons_intf.consumers):
            consumers.append(port)
        else:
            _get_end_consumer_rec(cons_intf, consumers)


def get_end_consumer(intf):
    consumers = []
    _get_end_consumer_rec(intf, consumers)
    return consumers


def _get_producer_rec(intf, producers):
    if isinstance(intf.producer, OutPort) or isinstance(
            intf.producer, GearOutPort):
        producers.append(intf.producer)
    else:
        _get_producer_rec(intf.producer, producers)


def get_producer(intf):
    producers = []
    _get_producer_rec(intf, producers)
    return producers


def find_target_prod(intf):
    # Who is producer gear?
    # prod_rtl_port = intf.producer
    end_p = get_producer(intf)
    if len(end_p) != 1:
        return None
    if isinstance(end_p[0], OutPort):
        prod_rtl_port = end_p[0]
        prod_rtl_node = prod_rtl_port.node
        prod_gear = prod_rtl_node.gear
    else:
        prod_gear = end_p[0].gear
    if len(prod_gear.child):
        if len(prod_gear.child) > 1:
            sim_log().warning(
                f'ActivityCosim: prod has more than one child. Setting on first.'
            )
        return prod_gear.child[0]
    else:
        return prod_gear


def find_target_cons(intf):
    # Who is consumer port? Case when not broadcast!
    # cons_rtl_port = intf.consumers[0]
    end_c = get_end_consumer(intf)
    if len(end_c) != 1:
        if len(end_c) > 1:
            sim_log().debug(f'Find target cons: found broadcast')
        return None
    cons_rtl_port = end_c[0]
    cons_rtl_node, port_id = cons_rtl_port.node, cons_rtl_port.index
    cons_gear = cons_rtl_node.gear
    cons_port = cons_gear.in_ports[port_id]

    return cons_port


def find_target_intf(gear_name, intf_name):
    gear_mod = find(gear_name)
    rtl_node = registry('RTLNodeMap')[gear_mod].node

    intf_name = intf_name[1:]  # spy name always starts with _
    for i in rtl_node.local_interfaces():
        if registry('SVGenMap')[i].basename == intf_name:
            return i


def set_waiting_edge(g, port):
    g.edge_map[port].set_color('blue')
    g.edge_map[port].set_penwidth(6)


def set_blocking_edge(g, port):
    g.edge_map[port].set_color('red')
    g.edge_map[port].set_penwidth(6)


def set_blocking_node(g, module):
    g.node_map[module].set_fillcolor('red')
    g.node_map[module].set_style('filled')


class ActivityReporter:
    def __init__(self, top, draw_graph=True):
        sim = registry('Simulator')
        sim.events['before_run'].append(self.before_run)
        sim.events['after_run'].append(self.after_run)
        self.blockers = {}
        self.draw_graph = draw_graph

    def intf_pull_start(self, intf):
        consumer = intf.producer
        producer = intf.in_queue.intf.consumers[0]
        self.blockers[consumer] = producer
        return True

    def intf_pull_done(self, intf):
        consumer = intf.producer
        del self.blockers[consumer]
        return True

    def before_run(self, sim):
        sim_map = registry('SimMap')

        for module, sim_gear in sim_map.items():
            for p in module.in_ports:
                p.consumer.events['pull_start'].append(self.intf_pull_start)
                p.consumer.events['pull_done'].append(self.intf_pull_done)

    def after_run(self, sim):

        if self.draw_graph:
            g = graph(
                outdir=registry('SimArtifactDir'),
                node_filter=lambda g: not g.child)
        else:
            g = None

        blocking_gears = set()
        cosim_name = None
        for sim_gear in sim.sim_gears:
            if isinstance(sim_gear, SimSocket):
                cosim_name = sim_gear.gear.name
                break
        if cosim_name:
            self.cosim_activity(g, cosim_name)
        self.sim_gears_activity(g, sim, blocking_gears)

        if self.draw_graph:
            outdir = registry('SimArtifactDir')
            g.graph.write_svg(os.path.join(outdir, 'proba.svg'))

        try:
            vcd_writer = registry('VCD')
        except KeyError:
            return

        with open(os.path.join(outdir, 'issue.gtkw'), 'w') as f:
            gtkw = GTKWSave(f)
            for module in blocking_gears:
                module_sav(gtkw, module, vcd_writer.vcd_vars)

    def sim_gears_activity(self, g, sim, blocking_gears):
        for sim_gear in sim.sim_gears:
            if isinstance(sim_gear, SimSocket):
                continue

            module = sim_gear.gear

            if self.draw_graph:
                g.node_map[module].set_style('filled')
                if sim_gear not in sim.done:
                    g.node_map[module].set_fillcolor('yellow')

            if module.definition == decoupler_din:
                if not module.queue.empty():
                    if self.draw_graph:
                        set_blocking_node(g, module)
                    blocking_gears.add(module)
                    sim_log().error(f'Data left in decoupler: {module.name}')

            for p in module.in_ports:
                q = p.get_queue()
                # print(f'{module.name}.{p.basename} queue empty: {q.empty()}')
                if q._unfinished_tasks:
                    src_port = q.intf.consumers[0]
                    if self.draw_graph:
                        set_blocking_edge(g, p)
                    blocking_gears.add(module)
                    sim_log().error(
                        f'{src_port.gear.name}.{src_port.basename} -> {module.name}.{p.basename} was not acknowledged'
                    )

                if p in self.blockers:
                    if self.draw_graph:
                        set_waiting_edge(g, p)
                    src_port = self.blockers[p]
                    sim_log().info(
                        f'{p.gear.name}.{p.basename} waiting on {src_port.gear.name}.{src_port.basename}'
                    )

    def cosim_activity(self, g, top_name):
        outdir = registry('SimArtifactDir')
        activity_path = os.path.join(outdir, 'activity.log')

        if not os.path.isfile(activity_path):
            return

        with open(activity_path, 'r') as log:
            for line in log:
                activity_name = line.rpartition(': ')[0]
                activity_name = activity_name.replace('top.dut.',
                                                      f'{top_name}/')
                activity_name = activity_name.rpartition('.')[0]
                activity_name = activity_name.replace('_i.', '/')
                gear_name, _, intf_name = activity_name.rpartition('/')

                # Const always has valid high
                const_regex = r'.*_const(?P<num>\d+)_s'
                const_regex_one = r'.*_const_s'
                if not (re.match(const_regex, intf_name)
                        or re.match(const_regex_one, intf_name)):
                    sim_log().error(
                        f'Cosim spy not acknowledged: {activity_name}')

                if self.draw_graph:
                    bc_regex = r'.*_bc_(?P<num>\d+).*'
                    if re.match(bc_regex, intf_name):
                        sim_log().debug(
                            f'Activity monitor cosim: bc not supported {activity_name}'
                        )
                        continue

                    intf = find_target_intf(gear_name, intf_name)
                    if intf is None:
                        sim_log().error(
                            f'Cannot find matching interface for {activity_name}'
                        )
                        continue
                    if intf.is_broadcast:
                        sim_log().debug(
                            f'Intf bc not supported {activity_name}')
                        continue

                    try:
                        prod_gear = find_target_prod(intf)
                        set_blocking_node(g, prod_gear)
                    except (KeyError, AttributeError):
                        sim_log().debug(
                            f'Cannot find node for {activity_name}')

                    try:
                        cons_port = find_target_cons(intf)
                        set_blocking_edge(g, cons_port)
                    except (KeyError, AttributeError):
                        sim_log().debug(
                            f'Cannot find edge for {activity_name}')



from .drv import drv, secdrv
from .verilator import SimVerilated
from .mon import mon
from .mon import delay_mon
from .seqr import seqr
from .scoreboard import scoreboard

__all__ = ['drv', 'mon', 'scoreboard', 'delay_mon', 'seqr', 'secdrv', 'SimVerilated']



from pygears.typing import Queue, Tuple, Unit, TypingNamespacePlugin


def flatten(dtype, lvl=1):
    if issubclass(dtype, Queue):
        return Queue[dtype[0], dtype.lvl - lvl]
    elif issubclass(dtype, Tuple):

        lvl -= 1
        out_args = []
        for a in dtype.args:
            if lvl:
                a = flatten(a, lvl)

            if issubclass(a, Tuple):
                out_args.extend(a.args)
            elif a != Unit:
                out_args.append(a)

        if len(out_args) == 0:
            return Unit
        elif len(out_args) == 1:
            return out_args[0]
        else:
            return Tuple[tuple(out_args)]
    else:
        return dtype


class TypeFlattenPlugin(TypingNamespacePlugin):
    @classmethod
    def bind(cls):
        cls.registry['TypeArithNamespace']['flatten'] = flatten



import ctypes
import os
import random
import re

import jinja2

from pygears.sim import sim_log
from pygears.sim.extens.rand_base import RandBase
from pygears.typing.visitor import TypingVisitorBase
from pygears.util.fileio import save_file


class SCVCompileError(Exception):
    pass


def file_changed(fn, new_contents):
    with open(fn) as f:
        data = f.read()

    return data != new_contents


def rebuild_needed(outdir, name, new_contents):
    lib_fn = os.path.join(outdir, name)
    cpp_fn = os.path.join(outdir, f'{name}.cpp')
    if not os.path.exists(lib_fn):
        return True

    return file_changed(cpp_fn, new_contents)


class SCVTypeSeqVisitor(TypingVisitorBase):
    def __init__(self, scvlib=None):
        self.scvlib = scvlib
        self.context = ''

        if self.scvlib:
            self.scvlib.init()
            self.scvlib.next()

    def visit(self, dtype, field=''):
        if field:
            '_'.join([self.context, field])
        val = super().visit(dtype, field)

        self.context.rsplit('_', 1)[0]
        return val

    def visit_queue(self, dtype, field):
        qlen = random.randrange(1, 3)
        return [self.visit(dtype[:-1]) for i in range(qlen)]

    def visit_tuple(self, dtype, field):
        return tuple(self.visit(d, f) for d, f in zip(dtype, dtype.fields))

    def visit_uint(self, dtype, field):
        scv_var_func = getattr(self.scvlib, f'get_{field}', None)
        if scv_var_func:
            return scv_var_func()
        else:
            return random.randrange(0, 2**(int(dtype)) - 1)

    def visit_int(self, dtype, field):
        scv_var_func = getattr(self.scvlib, f'get_{field}', None)
        if scv_var_func:
            return scv_var_func()
        else:
            return random.randrange(-2**(int(dtype) - 1),
                                    2**(int(dtype) - 1) - 1)


class SCVTypeVisitor(TypingVisitorBase):
    def __init__(self):
        self.cvars = {}
        self.depth = -1

    def visit(self, type_, field):
        self.depth += 1
        type_declaration = super().visit(type_, field)
        self.depth -= 1
        return type_declaration

    def visit_int(self, type_, field):
        self.cvars[field] = (f'sc_int<{int(type_)}>', int(type_))

    def visit_uint(self, type_, field):
        if (int(type_) != 0):
            self.cvars[field] = (f'sc_uint<{int(type_)}>', int(type_))

    def visit_unit(self, type_, field):
        return None

    def visit_union(self, type_, field):
        pass

    def visit_queue(self, type_, field):
        pass

    def visit_tuple(self, type_, field):
        for t, f in zip(type_.args, type_.fields):

            if field:
                self.visit(t, f'field_{f}')
            else:
                self.visit(t, f)

    def visit_array(self, type_, field):
        pass


class SCVConstraints:
    def __init__(self,
                 name='dlft',
                 dtype=None,
                 cvars={},
                 cons=[],
                 cls='scv_constraint_base'):
        self.name = name
        self.dtype = dtype
        self.cvars = cvars.copy()
        self.cons = cons.copy()
        self.cls = cls

    def add_var(self, name, dtype):
        v = SCVTypeVisitor()
        v.visit(dtype, name)
        self.cvars.update(v.cvars)


class SCVRand(RandBase):
    def create_type_cons(self, desc={}):
        v = SCVTypeVisitor()
        v.visit(desc['dtype'], desc['name'])
        tcons = SCVConstraints(
            name=desc['name'],
            dtype=desc['dtype'],
            cvars=v.cvars,
            cons=desc['cons'],
            cls=desc['cls'])
        for name, dtype in desc['params'].items():
            tcons.add_var(name, dtype)

        return tcons

    def before_setup(self, sim):
        self.create_scv()

    def create_scv(self):
        for constraint in self.constraints:
            jenv = jinja2.Environment(trim_blocks=True, lstrip_blocks=True)
            jenv.globals.update(int=int)
            jenv.loader = jinja2.FileSystemLoader([os.path.dirname(__file__)])

            def add_parens_for_vars(cons, cvars):
                def add_parens(matchobj):
                    if matchobj.group(0) in cvars:
                        return f'{matchobj.group(0)}()'
                    else:
                        return matchobj.group(0)

                return re.sub('([^\d\W]\w*)', add_parens, cons)

            scv_cons = [
                add_parens_for_vars(c, constraint.cvars)
                for c in constraint.cons
            ]
            context = {
                'vars': constraint.cvars,
                'constraints': scv_cons,
                'base_cls': constraint.cls
            }
            c = jenv.get_template('scv_wrap.j2').render(context)
            save_file(f'{constraint.name}.cpp', self.outdir, c)

    def get_rand(self, name):
        ret = os.system(
            f"cd {self.outdir}; g++ -fpic -shared -I. -I$SYSTEMC_INCLUDE -I$SCV_INCLUDE -Wall -Wformat -O2 $SCV_LIBDIR/libscv.so -L$SYSTEMC_LIBRID $SYSTEMC_LIBDIR/libsystemc.so -lpthread -pthread -Wl,-rpath -Wl,$SCV_LIBDIR -Wl,-rpath -Wl,$SYSTEMC_LIBDIR -o {name} {name}.cpp"
        )

        if ret != 0:
            sim_log().warning(
                'Possible reason for exception: Queues not supported in SCV')
            raise SCVCompileError(
                f'Constrained random library compilation failed for module {name}.'
            )

        scvlib = ctypes.CDLL(os.path.join(self.outdir, name))
        t = self.get_dtype_by_name(name)
        return SCVTypeSeqVisitor(scvlib).visit(t, name)



from pygears import gear, module
from pygears.core.log import gear_log
from pygears.typing import Queue, Uint


@gear
async def chop(din: Queue['data_t'], cfg: Uint['w_cfg']) -> Queue['data_t', 2]:

    i = 0
    val = din.dtype((0, 0))

    async with cfg as size:
        while (val.eot == 0):
            i += 1
            async with din as val:
                dout = module().tout((val.data, val.eot or (i % size == 0),
                                      val.eot))
                gear_log().debug(f'Chop yielding {dout}')
                yield dout



class Port:
    def __init__(self,
                 node,
                 index,
                 basename,
                 producer=None,
                 consumer=None,
                 dtype=None):
        self.node = node
        self.index = index
        self.producer = producer
        self.consumer = consumer
        self.basename = basename
        self.dtype = dtype


class InPort(Port):
    pass


class OutPort(Port):
    pass



from pygears.typing import Queue, Uint
from pygears import gear


@gear
def clip(din: Queue['T'], cfg: Uint, *, clip_stop=0) -> Queue['T']:
    ''' Clips the input transaction into two separate transactions by
sending eot after a given number of data has passed (specified by
configuration). The second eot is passed from input.

    clip_stop -- stop after first eot is sent
    '''
    pass



import ctypes
from math import ceil
from pygears.typing_common.codec import code, decode
from pygears.sim.modules.cosim_base import CosimNoData


class IdleIteration(Exception):
    pass


class CGet:
    def __init__(self, verilib, name, dtype):
        self.c_get_api = getattr(verilib, f'get_{name}', None)

        self.dtype = dtype
        self.width = int(dtype)

        if self.width > 64:
            self.c_dtype = ctypes.c_uint * (ceil(self.width / 32))
        elif self.width > 32:
            self.c_dtype = ctypes.c_ulonglong
        else:
            self.c_dtype = ctypes.c_uint

    def from_c_data(self, data):
        dout = 0
        for d in reversed(list(data)):
            dout <<= 32
            dout |= d

        return dout

    def get(self):
        self.c_get_api(self.dout)
        return self.from_c_data(self.dout)


class CDrv:
    def __init__(self, verilib, port):
        self.verilib = verilib
        self.port = port

        self.data_posted = False
        self.done = False
        self.width = int(self.port.dtype)
        self.c_set_api = getattr(verilib, f'set_{port.basename}', None)
        self.c_get_api = getattr(verilib, f'get_{port.basename}', None)

        if self.width > 64:
            self.c_dtype = ctypes.c_uint * (ceil(self.width / 32))
        elif self.width > 32:
            self.c_dtype = ctypes.c_ulonglong
        else:
            self.c_dtype = ctypes.c_uint


class CInputDrv(CDrv):
    def __init__(self, verilib, port):
        super().__init__(verilib, port)
        self.c_set_api.argtypes = (self.c_dtype, ctypes.c_uint)

    def close(self):
        pass

    def to_c_data(self, data):
        if self.width > 64:

            def dgen(data):
                for i in range(ceil(self.width / 32)):
                    yield data & 0xffffffff
                    data >>= 32

            return self.c_dtype(*list(dgen(data)))
        else:
            return self.c_dtype(data)

    def empty(self):
        return self.seq.empty()

    def send(self, data):
        self.c_set_api(self.to_c_data(code(self.port.dtype, data)), 1)

    def ready(self):
        return self.c_get_api()

    def reset(self):
        self.c_set_api(self.to_c_data(0), 0)


class COutputDrv(CDrv):
    def __init__(self, verilib, port):
        super().__init__(verilib, port)
        if self.width <= 64:
            self.c_dtype = self.c_dtype * 1
            self.c_get_api.argtypes = (self.c_dtype, )

        self.active = False
        self.dout = self.c_dtype()

    def from_c_data(self, data):
        dout = 0
        for d in reversed(list(data)):
            dout <<= 32
            dout |= d

        return dout

    def reset(self):
        self.c_set_api(0)

    def ack(self):
        self.c_set_api(1)

    def read(self):
        self.active = self.c_get_api(self.dout)
        # print(
        #     f'{self.port.basename}: {self.active}, {self.from_c_data(self.dout)}'
        # )
        if self.active:
            return decode(self.port.dtype, self.from_c_data(self.dout))
        else:
            raise CosimNoData



import inspect

from pygears import gear, GearDone, alternative
from pygears.sim import clk, sim_log
from pygears.typing import TLM
from pygears.util.utils import quiter
from pygears.sim.utils import SimDelay

# from pygears.sim.extens.svrand import get_rand_data


class TypingYieldVisitorBase:
    def visit(self, data, dtype):
        visit_func_name = f'visit_{dtype.__name__.lower()}'

        visit_func = getattr(self, visit_func_name, self.visit_default)
        ret = visit_func(data, dtype)
        if inspect.isgenerator(ret):
            yield from ret
        else:
            yield ret

    def visit_default(self, data, dtype):
        yield data


class TypeDrvVisitor(TypingYieldVisitorBase):
    def visit_queue(self, data, dtype):
        for (i, d), eot in quiter(enumerate(data)):
            for ret in self.visit(d, dtype[:-1]):
                if dtype.lvl == 1:
                    yield (ret, eot)
                else:
                    yield (ret[0], *ret[1:], eot)


@gear
async def drv(*, t, seq) -> b't':
    for val in seq:
        if type(val) == t:
            yield val
        else:
            for d in TypeDrvVisitor().visit(val, t):
                yield t(d)

    raise GearDone


@gear
async def secdrv(seqin, *, t) -> b't':
    async with seqin as seq:
        for val in seq:
            if type(val) == t:
                yield val
            else:
                for d in TypeDrvVisitor().visit(val, t):
                    yield t(d)


# @gear
# async def drv_rand_queue(*, tout, data_func,
#                          eot_con_name='din_eot') -> b'tout':
#     eot = 0
#     while eot != int('1' * tout.lvl, 2):
#         eot = get_rand_data(eot_con_name)
#         yield tout((data_func(), *eot))

#     raise GearDone



from pygears.core.gear import alternative, gear
from pygears.typing import Uint
from pygears.util.hof import oper_tree


@gear(enablement=b'len(din) == 2')
def eq(*din,
       din0_signed=b'typeof(din0, Int)',
       din1_signed=b'typeof(din1, Int)') -> Uint[1]:
    pass


@alternative(eq)
@gear
def eq_vararg(*din, enablement=b'len(din) > 2') -> Uint[1]:
    return oper_tree(din, eq)



def quiter(iterable):
    """Pass through all values from the given iterable, augmented by the
    information if there are more values to come after the current one
    (False), or if it is the last value (True).
    """
    # Get an iterator and pull the first value.
    it = iter(iterable)
    last = next(it)
    # Run the iterator to exhaustion (starting from the second value).
    for val in it:
        # Report the *previous* value (more to come).
        yield last, False
        last = val
    # Report the last value.
    yield last, True


async def quiter_async(intf):

    teot = intf.dtype[1:]
    eot = 0

    while (eot != ((1 << len(teot)) - 1)):
        data = await intf.pull()
        eot = data[1:]

        yield data

        intf.ack()


class gather:
    def __init__(self, *din):
        self.din = din

    async def __aenter__(self):
        din_data = []
        for d in self.din:
            din_data.append(await d.pull())

        return tuple(din_data)

    async def __aexit__(self, exception_type, exception_value, traceback):
        for d in self.din:
            d.ack()



from pygears.core.hier_node import NamedHierNode, find_unique_names
from pygears.rtl.port import InPort, OutPort
from pygears.rtl.intf import RTLIntf


def is_in_subbranch(root, node):
    root_path = root.name.split('/')
    node_path = node.name.split('/')

    if len(root_path) > len(node_path):
        return False

    for r, n in zip(root_path, node_path):
        if r != n:
            return False
    else:
        return True


class RTLNode(NamedHierNode):
    def __init__(self, parent, name, params={}):
        super().__init__(name, parent)
        self.in_ports = []
        self.out_ports = []
        self.params = params.copy()

    def add_in_port(self, basename, producer=None, consumer=None, dtype=None):
        self.in_ports.append(
            InPort(
                node=self,
                index=len(self.in_ports),
                basename=basename,
                producer=producer,
                consumer=consumer,
                dtype=dtype))

        port_names = [p.basename for p in self.in_ports]
        for port, new_name in zip(self.in_ports, find_unique_names(port_names)):
            if new_name:
                port.basename = new_name

    def add_out_port(self, basename, producer=None, consumer=None, dtype=None):
        self.out_ports.append(
            OutPort(
                node=self,
                index=len(self.out_ports),
                basename=basename,
                producer=producer,
                consumer=consumer,
                dtype=dtype))

        port_names = [p.basename for p in self.out_ports]
        for port, new_name in zip(self.out_ports, find_unique_names(port_names)):
            if new_name:
                port.basename = new_name

    def bypass(self):
        if not (len(self.in_ports) == 1 and len(self.out_ports) == 1):
            raise Exception(
                'Can only bypass single input, single output modules')

        iin = self.in_ports[0].producer
        iout = self.out_ports[0].consumer
        self.remove()

        for port in iout.consumers.copy():
            iout.disconnect(port)
            iin.connect(port)

        iout.remove()

    def remove(self):
        for p in self.in_ports:
            if p.producer is not None:
                p.producer.disconnect(p)

        for p in self.out_ports:
            p.consumer.producer = None

        super().remove()

    @property
    def is_hierarchical(self):
        return any([isinstance(c, RTLNode) for c in self.child])

    def local_interfaces(self):
        for child in self.child:
            if isinstance(child, RTLIntf):
                yield child

    def local_modules(self):
        for child in self.child:
            if isinstance(child, RTLNode):
                yield child

    @property
    def consumers(self):
        consumers = []
        for p in self.out_ports:
            iout = p.consumer
            consumers.extend(iout.consumers)

        return consumers

    def channel_ports(self):
        # If this is a top level module, no ports need to be output further
        if self.parent is None:
            return

        for p in self.in_ports():
            if p['intf'].parent != self.parent and (
                    not self.parent.is_descendent(p['intf'].parent)):
                self.parent.in_port_make(p, self)

        for p in self.out_ports():
            consumers_at_same_level_or_sublevel = [
                is_in_subbranch(p['intf'].parent, c[0])
                for c in p['intf'].consumers
            ]
            if not all(consumers_at_same_level_or_sublevel) or (
                    not p['intf'].consumers):
                self.parent.out_port_make(p, self)



from setuptools import setup, find_packages
from setuptools.command.develop import develop
from setuptools.command.install import install
from setuptools.command.egg_info import egg_info
import os
import glob


def setup_home():
    os.makedirs(os.path.expanduser('~/.pygears'), exist_ok=True)
    os.makedirs(os.path.expanduser('~/.pygears/svlib'), exist_ok=True)


class PostEggInfoCommand(egg_info):
    def run(self):
        egg_info.run(self)
        setup_home()


class PostDevelopCommand(develop):
    """Post-installation for development mode."""

    def run(self):
        develop.run(self)
        setup_home()


class PostInstallCommand(install):
    """Post-installation for installation mode."""

    def run(self):
        install.run(self)
        setup_home()


def readme():
    with open('README.rst') as f:
        return f.read()


setup(
    name='pygears',
    version='0.1.1',
    description='Framework for functional hardware design approach',
    long_description=readme(),

    # The project's main homepage.
    url='https://www.pygears.org',
    # download_url = '',

    # Author details
    author='Bogdan Vukobratovic',
    author_email='bogdan.vukobratovic@gmail.com',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
    ],
    python_requires='>=3.6.0',
    package_data={'': ['*.j2', '*.sv']},
    include_package_data=True,
    keywords='functional hardware design Python simulator HDL ASIC FPGA Gears',
    install_requires=['jinja2>=2.10', 'pygears-tools'],
    packages=find_packages(exclude=['examples*', 'docs']),
    entry_points={
        'console_scripts': [
            'pywave = pygears.sim.extens.pywave:main',
        ],
    },
    cmdclass={
        'develop': PostDevelopCommand,
        'install': PostInstallCommand,
        'egg_info': PostEggInfoCommand,
    },
)



class SVGenSyncGuard:
    def __init__(self, name, size):
        self.size = size
        self.intf_ports = []
        self.basename = name

        for i in range(size):
            self.add_port('consumer', f'din{i}')
            self.add_port('producer', f'dout{i}')

    @property
    def sv_file_name(self):
        return self.basename + ".sv"

    def add_port(self, modport, name):
        self.intf_ports.append({'modport': modport, 'name': name})

    def get_module(self, template_env):
        context = {'module_name': self.basename, 'intfs': self.intf_ports}
        return template_env.render_local(__file__, "syncguard.j2", context)



from functools import reduce


def divide(seq, length):
    return [seq[i: i+length] for i in range(0, len(seq), length)]


def oper_tree(din, func):
    pairs = divide(din, 2)
    results = []
    for i in pairs:
        if(len(i) == 2):
            results.append(func(*i))
        else:
            results.append(i)

    return func(*results)


def oper_reduce(din, func, init=0):
    return reduce(func, din, init)



class TypingVisitorBase:
    def visit(self, type_, field=None, **kwds):
        visit_func_name = f'visit_{type_.__name__.lower()}'

        visit_func = getattr(self, visit_func_name, self.visit_default)

        return visit_func(type_, field, **kwds)

    def visit_union(self, type_, field, **kwds):
        for t, f in zip(type_.types, type_.fields):
            self.visit(t, f)

    def visit_int(self, type_, field, **kwds):
        pass

    def visit_bool(self, type_, field, **kwds):
        pass

    def visit_uint(self, type_, field, **kwds):
        pass

    def visit_default(self, type_, field, **kwds):
        if hasattr(type_, 'fields'):
            return {
                f: self.visit(t, f, **kwds)
                for t, f in zip(type_, type_.fields)
            }
        else:
            try:
                return tuple(
                    self.visit(t, **kwds)
                    for t in type_)
            except TypeError:
                pass



from .sim_extend import SimExtend
import time


class SimTiming(SimExtend):
    def __init__(self, top, window=1000):
        super().__init__()
        self.window = window
        self.forward_times = {}
        self.back_times = {}
        self.total_forward_times = {}
        self.total_back_times = {}

    def before_run(self, sim):
        for sim_gear in sim.sim_gears:
            self.forward_times[sim_gear] = 0
            self.back_times[sim_gear] = 0
            self.total_forward_times[sim_gear] = 0
            self.total_back_times[sim_gear] = 0

    def before_timestep(self, sim, timestep):
        if (timestep % self.window) == 0:
            if timestep > 0:
                print(
                    f"-------------------- {timestep} ----------------------")

                sort = sorted(
                    ((sim_gear, self.forward_times[sim_gear],
                      self.back_times[sim_gear])
                     for sim_gear in sim.sim_gears),
                    key=lambda x: x[1] + x[2])

                print(
                    f'{"name":60} | {"total [ms]":10} | {"forward [us]":12} | {"back [us]":9}')
                print("-" * 100)
                for sim_gear, ft, bt in reversed(sort):
                    forward = int(ft / self.window * 1000000)
                    back = int(bt / self.window * 1000000)
                    total = int((self.total_forward_times[sim_gear] +
                                 self.total_back_times[sim_gear]) * 1000)
                    print(
                        f'{sim_gear.gear.name:60} | {total:10} | {forward:12} | {back:9}'
                    )
                print("-" * 100)

            for sim_gear in sim.sim_gears:
                self.forward_times[sim_gear] = 0
                self.back_times[sim_gear] = 0

        return True

    def before_call_forward(self, sim, sim_gear):
        self.start_time = time.time()
        return True

    def after_call_forward(self, sim, sim_gear):
        delta = time.time() - self.start_time
        self.forward_times[sim_gear] += delta
        self.total_forward_times[sim_gear] += delta
        return True

    def before_call_back(self, sim, sim_gear):
        self.start_time = time.time()
        return True

    def after_call_back(self, sim, sim_gear):
        delta = time.time() - self.start_time
        self.back_times[sim_gear] += delta
        self.total_back_times[sim_gear] += delta
        return True



import asyncio
import random

class Module:
    def __init__(self, in_ports, out_ports):
        pass

    def connect(self, port):
        if len(out_ports) == 1:
            self.out_ports[0].connect(port)

async def produce(queue, n):
    for x in range(n):
        # produce an item
        print('producing {}/{}'.format(x, n))
        # simulate i/o operation using sleep
        item = str(x)

        # put the item in the queue
        for q in queue:
            q.put_nowait(item)

        await asyncio.wait([q.join() for q in queue])

        print('all done')

        await asyncio.sleep(1)


async def consume(queue, cid):
    while True:
        # wait for an item from the producer
        item = await queue.get()

        # process the item
        print('{}: consuming {}...'.format(cid, item))
        # Notify the queue that the item has been processed
        queue.task_done()

        # simulate i/o operation using sleep
        await asyncio.sleep(2+cid)


async def run(n):
    queue = [asyncio.Queue(maxsize=1) for _ in range(2)]
    # schedule the consumer
    consumer = [
        asyncio.ensure_future(consume(q, i)) for i, q in enumerate(queue)
    ]

    # run the producer and wait for completion
    await produce(queue, n)
    # the consumer is still awaiting for an item, cancel it
    for c in consumer:
        c.cancel()


loop = asyncio.get_event_loop()
loop.run_until_complete(run(10))
loop.close()



from pygears.core.hier_node import NamedHierNode
from pygears.rtl.port import InPort, OutPort


class RTLIntf(NamedHierNode):
    def __init__(self, parent, dtype, producer=None, consumers=[]):
        super().__init__(basename=None, parent=parent)
        self.consumers = consumers.copy()
        self.producer = producer
        self.dtype = dtype

    @property
    def basename(self):
        producer_port = self.producer
        port_name = producer_port.basename

        if hasattr(self, 'var_name'):
            producer_name = self.var_name
        else:
            producer_name = producer_port.node.basename

        if isinstance(producer_port, InPort):
            return port_name
        elif ((not self.is_broadcast) and self.consumers
              and isinstance(self.consumers[0], OutPort)):
            return self.consumers[0].basename
        elif self.sole_intf:
            return f'{producer_name}_s'
        else:
            return f'{producer_name}_{port_name}_s'

    @property
    def outname(self):
        if self.is_broadcast:
            return f'{self.basename}_bc'
        else:
            return self.basename

    @property
    def sole_intf(self):
        if self.producer:
            return len(self.producer.node.out_ports) == 1
        else:
            return True

    @property
    def is_broadcast(self):
        return len(self.consumers) > 1

    @property
    def is_port_intf(self):
        if isinstance(self.producer, InPort):
            return True
        elif ((not self.is_broadcast) and self.consumers
              and isinstance(self.consumers[0], OutPort)):
            return True
        else:
            return False

    def disconnect(self, port):
        if port in self.consumers:
            self.consumers.remove(port)
            port.producer = None
        elif port == self.producer:
            port.consumer = None
            self.producer = None

    def connect(self, port):
        self.consumers.append(port)
        port.producer = self



import jinja2
import os
import ctypes
from pygears.util.fileio import save_file
from pygears import registry
from pygears.svgen import svgen
from pygears.sim.modules.cosim_base import CosimBase
from pygears.sim.c_drv import CInputDrv, COutputDrv
from pygears.sim import sim_log
import atexit


class VerilatorCompileError(Exception):
    pass


class SimVerilatorSynchro:
    def __init__(self, verilib):
        self.verilib = verilib

    def cycle(self):
        self.verilib.trig()

    def forward(self):
        self.verilib.eval()

    def back(self):
        self.verilib.back()


class SimVerilated(CosimBase):
    def __init__(self, gear):
        super().__init__(gear, timeout=100)
        self.name = gear.name[1:].replace('/', '_')
        self.outdir = os.path.abspath(
            os.path.join(registry('SimArtifactDir'), self.name))
        self.objdir = os.path.join(self.outdir, 'obj_dir')
        self.svnode = svgen(gear, outdir=self.outdir, wrapper=True)
        self.svmod = registry('SVGenMap')[self.svnode]
        self.wrap_name = f'wrap_{self.svmod.sv_module_name}'

    def setup(self):
        rebuild = True

        if rebuild:
            sim_log().info(f'Verilating...')
            self.build()
            sim_log().info(f'Done')

        self.verilib = ctypes.CDLL(
            os.path.join(self.objdir, f'V{self.wrap_name}'))

        self.finished = False
        atexit.register(self.finish)
        self.verilib.init()

        self.handlers = {}
        for p in self.gear.in_ports:
            self.handlers[p.basename] = CInputDrv(self.verilib, p)

        for p in self.gear.out_ports:
            self.handlers[p.basename] = COutputDrv(self.verilib, p)

        self.handlers[self.SYNCHRO_HANDLE_NAME] = SimVerilatorSynchro(
            self.verilib)

    def build(self):
        context = {
            'in_ports': self.svnode.in_ports,
            'out_ports': self.svnode.out_ports,
            'top_name': self.wrap_name,
            'tracing': True,
            'outdir': self.outdir
        }

        include = ' '.join([
            f'-I{os.path.abspath(p)}'
            for p in registry('SVGenSystemVerilogPaths')
        ])

        jenv = jinja2.Environment(trim_blocks=True, lstrip_blocks=True)
        jenv.globals.update(int=int)
        jenv.loader = jinja2.FileSystemLoader([os.path.dirname(__file__)])
        c = jenv.get_template('sim_veriwrap.j2').render(context)
        save_file('sim_main.cpp', self.outdir, c)

        ret = os.system(
            f"cd {self.outdir}; verilator -cc -CFLAGS -fpic -LDFLAGS -shared --exe {include} -clk clk --trace --trace-structs --top-module {self.wrap_name} {self.outdir}/*.sv dti.sv sim_main.cpp -Wno-fatal > verilate.log 2>&1"
        )

        if ret != 0:
            raise VerilatorCompileError(
                f'Verilator compile error: {ret}. '
                f'Please inspect "{self.outdir}/verilate.log"')

            # if not os.path.exists(self.objdir):
            #     raise VerilatorCompileError(
            #         f'Verilator compile error: {ret}. '
            #         f'Please inspect "{self.outdir}/verilate.log"')
            # else:
            #     sim_log().warning(
            #         f'Verilator compiled with warnings. '
            #         f'Please inspect "{self.outdir}/verilate.log"')

        ret = os.system(
            f"cd {self.objdir}; make -j -f V{self.wrap_name}.mk > make.log 2>&1"
        )

        if ret != 0:
            raise VerilatorCompileError(
                f'Verilator compile error: {ret}. '
                f'Please inspect "{self.outdir}/make.log"')

        sim_log().info(f'Verilator VCD dump to "{self.outdir}/vlt_dump.vcd"')

    def finish(self):
        if not self.finished:
            self.finished = True
            super().finish()
            self.verilib.final()



from pygears.svgen.inst import SVGenInstPlugin
from pygears.svgen.inst import svgen_inst
from pygears.rtl.gear import RTLGearHierVisitor, is_gear_instance
from pygears.svgen.util import svgen_visitor


@svgen_visitor
class RemoveTupleFlattenVisitor(RTLGearHierVisitor):
    def flatten_tuple(self, node):
        node.bypass()


class SVGenFlattenPlugin(SVGenInstPlugin):
    @classmethod
    def bind(cls):
        cls.registry['SVGenFlow'].insert(
            cls.registry['SVGenFlow'].index(svgen_inst),
            RemoveTupleFlattenVisitor)



from pygears import gear
from pygears.typing import Queue


@gear
def tr_cnt(din: Queue['TData'], cfg: 'TCfg') -> Queue['TData', 2]:
    '''Transaction counter: counts the input eots. Number of eots to count
    given with cfg. When sufficent transactions are seen, returns ready on cfg
    and sets higher eot on output

    din -- Queue
    cfg -- how many transactions to count
    dout -- Queue, lvl 2: lower list same as input, higher list shows that
    transactions were counted
    '''
    pass



from pygears.core.gear import gear
from pygears.typing import Integer, Int
from pygears.core.intf import IntfOperPlugin
from pygears import module


def neg_type(dtype):
    return Int[int(dtype)]


@gear(svgen={'svmod_fn': 'neg.sv'})
async def neg(din: Integer) -> b'neg_type(din)':

    async with din as d:
        yield module().tout(-d)


class NegIntfOperPlugin(IntfOperPlugin):
    @classmethod
    def bind(cls):
        cls.registry['IntfOperNamespace']['__neg__'] = neg



from pygears import gear, alternative, module
from pygears.typing import Queue
from pygears.typing_common.flatten import flatten as type_flatten
from operator import and_

from functools import reduce


def flatten_func(d, lvl):
    dout_lvl = d.lvl - lvl

    dout = [d[0]]
    if lvl > 1:
        dout.append(reduce(and_, d[1:lvl + 2]))

    if dout_lvl > 1:
        dout += list(d[lvl + 2:])

    if len(dout) == 1:
        dout = dout[0]

    return Queue[type(d)[0], dout_lvl](dout)


@gear
async def flatten(din: Queue['tdin', 'din_lvl'],
                  *,
                  lvl=1,
                  dout_lvl=b'din_lvl - lvl') -> b'Queue[tdin, dout_lvl]':

    async with din as d:
        yield flatten_func(d, lvl)


@alternative(flatten)
@gear(enablement=b'issubclass(din, Tuple)')
def flatten_tuple(din, *, lvl=1):
    return din | type_flatten(din.dtype, lvl)
